/* ---------------------------------------------------------------------------
 *  Cirrus Logic fb operators
 * --------------------------------------------------------------------------
 *
 *  Copyright (C) 1997, The Linux-GGI Project
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 *
 *  This code was derived from the following sources of information:
 *  [1]     Xfree 86 VGA driver
 *  [2]     Zen of Graphics Programming (Michael Abrash)
 */


static void cirrus_set_origin(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	IS_CRITICAL;
	static ggi_uint last_xdot = -1;
	static ggi_uint last_ydot = -1;
	static ggi_uint old_offset = -1;
	register ggi_uint xpix=0,xdot=0;
	register ggi_uint ypix=0,ydot=0;
	register ggi_uint offset;
	ggi_uint fb_width=0;

	switch (dpy->mode.request.graphtype) {
		case GT_TEXT16:
			xpix = x / dpy->mode.request.dpp.x;
			xdot = x % dpy->mode.request.dpp.x;
			ypix = y / dpy->mode.request.dpp.y;
			ydot = y % dpy->mode.request.dpp.y;
			fb_width = dpy->mode.request.virt.x * 2;
			break;
		case GT_1BIT:
		case GT_4BIT:
			xpix = x / 8;
			xdot = x % 8;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x / 8;
			break;
		case GT_8BIT:
			xpix = x;
			xdot = 0;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x;
			break;	
		case GT_15BIT:
		case GT_16BIT:
			xpix = x;
			xdot = 0;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*2;
			break;
		case GT_24BIT:
			xpix = x;
			xdot = 0;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*3;
			break;
		case GT_32BIT:
			xpix = x;
			xdot = 0;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*4;
			break;
		default:
			INTERNALERROR;
	}
		
 	offset = xpix + ypix * fb_width;

	ENTER_CRITICAL;

	if (old_offset != offset) {

		outbCRT(offset >> 8, CRT_STARTADDR_H);
		outbCRT(offset, CRT_STARTADDR_L);
		outbCRT((inbCRT(0x1B) & 0xF2) |
			((offset>>16) & 0x01) |
			((offset>>15) & 0x0C) , 0x1B);
		old_offset = offset;
	}

	if (last_ydot != ydot) {
		outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
		last_ydot = ydot;
	}

	if (last_xdot != xdot) {

		register ggi_uint hpan;

		if (dpy->mode.request.graphtype == GT_TEXT16 &&
		    dpy->mode.request.dpp.x == 9) {
			
			hpan = (xdot == 0) ? 0x08 : (xdot-1) & AR13_DATA_MASK;
		} else {
			
			hpan = xdot & AR13_DATA_MASK;
		}
		WAIT_HVRETRACE();
		outbATC(hpan, ATC_HPAN);
		outb( ATCI_ENABLEDISPLAY, ATCI);
		last_xdot = xdot;
	}
	
	LEAVE_CRITICAL;
}

static void cirrus_set_split(struct kgi_display *dpy, int line)
{
	outbCRT(( line        & 0xFF) , 0x18);

	outbCRT((inbCRT(0x07) & 0xEF) | 
		((line  >> 4) & 0x10) , 0x07);

	outbCRT((inbCRT(0x09) & 0xBF) | 
		((line >>  3) & 0x40) , 0x09);
}

#if 0
static void cirrus_set_pointer_shape(struct kgi_display *dpy,struct kgi_pointer64x64 *shape)
{
	switch(shape->mode)
	{
	case pmWindows:
	case pmX11:
	case pmThreeColor:
	default:
	}
}

static void cirrus_show_pointer(struct kgi_display *dpy,int x,int y)
{
	outw((x<<5)|0x10,0x3C4);
	outw((y<<5)|0x11,0x3C4);
	outw(0x8312,0x3C4);
}

static void cirrus_hide_pointer(struct kgi_display *dpy)
{
	outbSEQ(inbSEQ(0x12)&~1,0x12);
}

#define cirrus_undo_pointer NULL;
#endif
