/* --------------------------------------------------------------------------
**	Cirrus Logic Laguna 546x IO functions
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997 Chris Lattner
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** --------------------------------------------------------------------------
**	MAINTAINER	Chris_Lattner
*/

#include "Cirrus/546x.h"  /* Make sure the register defines are available */

/* Number of registers to save by SaveState */
#define NoSEQRegs       0x07	/* (Number-1) in fact */
#define NoGRCRegs       0x08	/* (Number-1) in fact */
#define NoATCRegs       0x14	/* (Number-1) in fact */
#define NoMMIORegs      0x20	/* (Number-1) in fact */

struct StateTable
{
	uint8  SEQ[NoSEQRegs+1];
	uint8  GRC[NoGRCRegs+1];
	uint8  ATC[NoATCRegs+1];
	uint8  MMIO[NoMMIORegs+1];
};

#if defined(DEBUG)
/* ----------------------------------------------------------------------------
**	Debug Functions
** ----------------------------------------------------------------------------
*/
static void dump_state(struct StateTable *state)
{
  int i;
  for(i = 0; i <= NoSEQRegs; i++)
   { if(!(i & 0xF)) printk("\nSEQ:%.2x: ",i); printk("%.2x, ",state->SEQ[i]); }
  for(i = 0; i <= NoGRCRegs; i++)
   { if(!(i & 0xF)) printk("\nGRC:%.2x: ",i); printk("%.2x, ",state->GRC[i]); }
  for(i = 0; i <= NoATCRegs; i++)
   { if(!(i & 0xF)) printk("\nATC:%.2x: ",i); printk("%.2x, ",state->ATC[i]); }
  for(i = 0; i <= NoMMIORegs; i++)
   { if(!(i & 0xf)) printk("\nMMIO:%.2x: ",i); printk("%.2x, ",(uint8)state->MMIO[i]); }
  printk("\n");
}
#endif

/* ----------------------------------------------------------------------------
**	I/O Functions
** ----------------------------------------------------------------------------
*/

/*static*/ int CLIO_in8(int Register)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		return mem_in8(linIOBuffer.baseptr+Register);

	} else if (RegSet == IOSROrValue) {

		io_out8(Register, SRX_INDEX);
		return io_in8(SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out8(Register, GRX_INDEX);
		return io_in8(GRD_DATA);

	} else if (RegSet == IOATCOrValue) {
		uint8 foo;

		io_in8(STAT_INPUT_STATUS_1);	/* reset AFF */
		io_out8(Register, ARX_INDEX);	/* set index */
		foo = io_in8(ARD_DATA_READ);		/* read value */
		io_in8(STAT_INPUT_STATUS_1);	/* reset AFF */
		io_out8(ARX_VIDEO_ENABLE, ARX_INDEX);
		return foo;

	} else {

		return io_in8(Register);
	}
}

/*static*/ void CLIO_out8(int Register, int value)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		mem_out8(value, linIOBuffer.baseptr+Register); 

	} else if (RegSet == IOSROrValue) {

		io_out8(Register, SRX_INDEX);
		io_out8(value, SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out8(Register, GRX_INDEX);
		io_out8(value, GRD_DATA);

	} else if (RegSet == IOATCOrValue) {

		io_in8(STAT_INPUT_STATUS_1);	/* reset AFF */
		io_out8(Register, ARX_INDEX);	/* set index */
		io_out8(value, ARD_DATA_WRITE);		/* write value */
		io_in8(STAT_INPUT_STATUS_1);	/* reset AFF */
		io_out8(ARX_VIDEO_ENABLE, ARX_INDEX);

	} else {

		io_out8(value, Register);
	}
    
	return;
}

static void CLIO_ins8(int Register, void *buf, int count)
{
	int i;
	for (i = 0; i < count; i++) {

		*((uint8 *)buf+i) = CLIO_in8(Register);
	}
}

static void CLIO_outs8(int Register, void *buf, int count)
{
	int i;
	for (i = 0; i < count; i++) {

		CLIO_out8(Register, *((uint8 *)buf+i));
	}
}

#warning CLIO_in/out_16/32 are to be checked -- ortalo
/* static */ int CLIO_in16(int Register)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		return mem_in16(linIOBuffer.baseptr+Register);

	} else {
		NOTICE("Non MMIO 16 bits in register !!");
		return io_in16(Register);
	}
}

/* static */ void CLIO_out16(int Register, int value)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		mem_out16(value, linIOBuffer.baseptr+Register); 

	} else {
		NOTICE("Non MMIO 16 bits out register !!");
		io_out16(value, Register);
	}
    
	return;
}
/*static*/ int CLIO_in32(int Register)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		return mem_in32(linIOBuffer.baseptr+Register);

	} else {
		NOTICE("Non MMIO 32 bits in register !!");
		return io_in32(Register);
	}
}

/*static*/ void CLIO_out32(int Register, int value)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		mem_out32(value, linIOBuffer.baseptr+Register); 

	} else {
		NOTICE("Non MMIO 32 bits out register !!");
		io_out32(value, Register);
	}
    
	return;
}
#if 0
static int CLIO_in16(int Register)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		return mem_in16(linIOBuffer.baseptr+Register);

	} else if (RegSet == IOSROrValue) {

		io_out16(Register, SRX_INDEX);
		return io_in16(SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out16(Register, GRX_INDEX);
		return io_in16(GRD_DATA);

	} else if (RegSet == IOATCOrValue) {

		io_out16(Register, ARX_INDEX);
		return io_in16(ARD_DATA_READ);

	} else {

		return io_in16(Register);
	}
}

static void CLIO_out16(int Register, int value)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		mem_out16(value, linIOBuffer.baseptr+Register); 

	} else if (RegSet == IOSROrValue) {

		io_out8(Register, SRX_INDEX);
		io_out8(value, SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out8(Register, GRX_INDEX);
		io_out8(value, GRD_DATA);

	} else if (RegSet == IOATCOrValue) {

		io_out8(Register, ARX_INDEX);
		io_out8(value, ARD_DATA_WRITE);

	} else {

		io_out16(value, Register);
	}
}

static int CLIO_in32(int Register)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		return mem_in32(linIOBuffer.baseptr+Register);

	} else if (RegSet == IOSROrValue) {

		io_out8(Register, SRX_INDEX);
		return io_in8(SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out8(Register, GRX_INDEX);
		return io_in8(GRD_DATA);

	} else if (RegSet == IOATCOrValue) {

		io_out8(Register, ARX_INDEX);
		return io_in8(ARD_DATA_READ);

	} else {

		return io_in32(Register);
	}
}

static void CLIO_out32(int Register, int value)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
  
	if (RegSet == IOMMOrValue) {

		mem_out32(value, linIOBuffer.baseptr+Register); 

	} else if (RegSet == IOSROrValue) {

		io_out8(Register, SRX_INDEX);
		io_out8(value, SRD_DATA);

	} else if (RegSet == IOGROrValue) {

		io_out8(Register, GRX_INDEX);
		io_out8(value, GRD_DATA);

	} else if (RegSet == IOATCOrValue) {

		io_out8(Register, ARX_INDEX);
		io_out8(value, ARD_DATA_WRITE);

	} else {

		io_out32(value, Register);
	}
}
#endif

static void CLIO_SetStateRegister(struct StateTable *state, int Register, 
	int Value, int size)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
	if (RegSet == IOMMOrValue) {

		PASSERT(size==1, "Saved MMIO for 1 bytes CRT regs, but i=%d",
			size);
		state->MMIO[Register >> 2] = (uint8)Value;

	} else if (RegSet == IOSROrValue) {

		state->SEQ[Register] = (uint8)Value;

	} else if (RegSet == IOGROrValue) {

		state->GRC[Register] = (uint8)Value;

	} else if (RegSet == IOATCOrValue) {

		state->ATC[Register] = (uint8)Value;

	} else {

		INTERNAL_ERROR;
	}
}

static uint32 CLIO_GetStateRegister(struct StateTable *state, int Register, 
	int size)
{
	int RegSet = Register & IOAllOrValues;

	Register &= IOMask;
	if (RegSet == IOMMOrValue) {

		PASSERT(size==1, "Saved MMIO for 1 bytes CRT regs, but i=%d",
			size);
		return state->MMIO[Register >> 2];

	} else if (RegSet == IOSROrValue) {

		return state->SEQ[Register];

	} else if (RegSet == IOGROrValue) {

		return state->GRC[Register];

	} else if (RegSet == IOATCOrValue) {

		return state->ATC[Register];

	} else {

		INTERNAL_ERROR;
	}
	return 0;
}

/* --------------------------------------------------------------------------
**	DAC IO Functions
** --------------------------------------------------------------------------
*/

static unsigned short dac_port[4] =
{
	DAC_PALETTE_ADDRESS_WRITE,
	DAC_PIXEL_DATA,
	DAC_PIXEL_MASK,
	DAC_PALETTE_ADDRESS_READ
};

void DAC_out8(struct kgi_display *dpy, uint8 value, uint8 dac_reg)
{
	ASSERT(dac_reg < 4);
	CLIO_out8(dac_port[dac_reg & 0x03], value);
}

uint8 DAC_in8(struct kgi_display *dpy, uint8 dac_reg)
{
	ASSERT(dac_reg < 4);
	return CLIO_in8(dac_port[dac_reg & 0x03]);
}

void DAC_ins8(struct kgi_display *dpy, uint8 dac_reg, void *buf, ggi_uint count)
{
	ASSERT(dac_reg < 4);
	CLIO_ins8(dac_port[dac_reg & 0x03], buf, count);
}

void DAC_outs8(struct kgi_display *dpy, uint8 dac_reg, void *buf, ggi_uint count)
{
	ASSERT(dac_reg < 4);
	CLIO_outs8(dac_port[dac_reg & 0x03], buf, count);
}

/* --------------------------------------------------------------------------
**	Wait functions
** --------------------------------------------------------------------------
*/
static inline void wait_display_off(struct kgi_display *dpy)
{
	while (~CLIO_in8(STAT_INPUT_STATUS_1) & STAT_DISPLAY_ENABLE);
}
static inline void wait_vertical_retrace(struct kgi_display *dpy)
{
	while (~CLIO_in8(STAT_INPUT_STATUS_1) & STAT_VERTICAL_RETRACE);
}

static inline void cl546x_unlock(void) {
	/* We must make sure that the write protection on CR0-7 of
	** CR11[7] is cleared, first.
	*/
	CLIO_out8(CR11_VERTICAL_SYNC_END,
		CLIO_in8(CR11_VERTICAL_SYNC_END) & ~CR11_WRITE_PROTECT_CR0_7);
	/* We must also make sure that the write protection on CR10-11 by
	** CR3 is disabled also.
	*/
	CLIO_out8(CR3_HORIZONTAL_BLANKING_END,
		CLIO_in8(CR3_HORIZONTAL_BLANKING_END) | CR3_COMPATIBLE_READ);
	/* Maybe we should check that extended display control is
	** enabled also before the programming.
	*/
	CLIO_out8(CR1B_EXTENDED_DISPLAY_CONTROL,
		CLIO_in8(CR1B_EXTENDED_DISPLAY_CONTROL) |
		CR1B_ENABLE_BLANK_END_EXTENSIONS |
		CR1B_ENABLE_EXTENDED_ADDRESS_WRAP);
	return;
}

/* --------------------------------------------------------------------------
**	Clock IO Functions
** --------------------------------------------------------------------------
*/

void CLK_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
	ASSERT(reg == 0);

	switch (value) {

	case 0: 
		value = MISC_VCLK0; /* vclk 0 usually 25 MHz */
		break;
	case 1:
		value = MISC_VCLK1; /* vclk 0 usually 28 MHz */
		break;
	default:
		INTERNAL_ERROR;
		return;
	}

	CLIO_out8(MISC_OUTPUT,
		(CLIO_in8(MISC_INPUT) & (~MISC_VCLK_MASK)) | value);
}

uint8 CLK_in8(struct kgi_display *dpy, uint8 reg)
{
	ASSERT(reg == 0);
	 return (CLIO_in8(MISC_INPUT) & MISC_VCLK_MASK) >> 2;
}

/* --------------------------------------------------------------------------
**	State Saving Routines
** --------------------------------------------------------------------------
*/

static void SaveMMIO(uint8 *buf)
{
	int i;
	NOTICE2("MMIO-Saving only CRT&MISC bytes of linIOBuffer");

	for (i=0; i<=NoMMIORegs; i++)
	{
		*(buf++) = mem_in8(linIOBuffer.baseptr + 4*i);
	}
}

static void RestoreMMIO(uint8 *buf)
{  
	int i;
	NOTICE("MMIO-Restoring only CRT&MISC bytes of linIOBuffer");

	cl546x_unlock();

	for (i=0; i<=NoMMIORegs; i++)
	{
		if ( (4*i) == (MISC_OUTPUT & IOMask) ) {
			/* We have to avoid to reset the clock !! */
			uint8 new_value;
			new_value = mem_in8((linIOBuffer.baseptr + 4*i))
				& MISC_VCLK_MASK;
			new_value |= (*(buf++)) & ~MISC_VCLK_MASK;
			mem_out8( new_value, (linIOBuffer.baseptr + 4*i));
			NOTICE("Special MISC reg restore");
			DEBUG1("Setting MISC reg preserving clock:%.2x",new_value);
		} else {
			mem_out8(*(buf++),(linIOBuffer.baseptr + 4*i));
		}
	}
}

static void SaveSEQ(uint8 *buf)
{
	int i;
	for(i = 0; i <= NoSEQRegs; i++) {

		*(buf++) = CLIO_in8(IOSRBaseRegister+i);
	}
}

static void RestoreSEQ(uint8 *buf)
{
	int i;

	for(i = 0; i <= NoSEQRegs; i++) {

		CLIO_out8(IOSRBaseRegister+i, *(buf++));
	}
}

static void SaveGRC(uint8 *buf)
{
	int i;

	for(i = 0; i <= NoGRCRegs; i++) {

		*(buf++) = CLIO_in8(IOGRBaseRegister+i);
	}
}

static void RestoreGRC(uint8 *buf)
{
	int i;

	for(i = 0; i <= NoGRCRegs; i++) {

		CLIO_out8(IOGRBaseRegister+i, *(buf++));
	}
}

static void SaveATC(uint8 *buf)
{
	int i;

	for(i = 0; i <= NoATCRegs; i++) {

		*(buf++) = CLIO_in8(IOATCBaseRegister+i);
	}
}

static void RestoreATC(uint8 *buf)
{
	int i;

	for(i = 0; i <= NoATCRegs; i++) {

		CLIO_out8(IOATCBaseRegister+i, *(buf++));
	}
}

static void SaveState(struct StateTable *state)
{
	NOTICE1("Entering SaveState");
	SaveSEQ(state->SEQ);
	SaveGRC(state->GRC);
	SaveATC(state->ATC);
	SaveMMIO(state->MMIO);
	NOTICE1("Done SaveState");
	NOTICE1("Manual dump of CRT registers");
	TRACE1(dump_state(state));
}

static void RestoreState(struct StateTable *state)
{
	struct StateTable test;
	NOTICE1("Restoring state:");
	TRACE1(dump_state(state));
	RestoreSEQ(state->SEQ);
	RestoreGRC(state->GRC);
	RestoreATC(state->ATC);
	RestoreMMIO(state->MMIO);
	NOTICE1("Re-reading state for test");
	SaveState(&test);
}

