/* ---------------------------------------------------------------------------
 *  IBM (vga) text16 operators - enhanced!
 * --------------------------------------------------------------------------
 *
 *  Copyright (C) 1997      Jason McMullan, [jmcc@cs.cmu.edu]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 *
 *  $Log: vga_text16_enh.inc,v $
 *  Revision 1.22  1998/02/13 08:04:04  bjulin
 *  Fixed HW cursor and mouse bounds checking
 *
 *  Revision 1.21  1998/02/09 15:26:26  grimrath
 *  New KGI driver Makefile system.  Include faster dependencies, true linking and
 *  a predefined driver constant for checking with #ifdef ... #endif
 *
 *  VGA, the Mystique driver and kernel files have been adapted to the
 *  text <-> dpp change.
 *
 *  Revision 1.20  1998/01/30 18:38:08  ursetto
 *  Fixed screen corruption in text16_font_enter.
 *
 *  Revision 1.19  1998/01/22 02:03:53  bjulin
 *  Added missing splitline code for show_cursor when defined CURSOR_HW
 *
 *  Revision 1.18  1997/12/14 21:55:42  becka
 *  uint/sint namespace fixed ... please test ...
 *
 *  Revision 1.17  1997/12/13 19:32:21  becka
 *  VGA fixes for ->private. autoconfig more PCI detection.
 *
 *  Revision 1.16  1997/12/09 10:11:39  seeger
 *  Global cleanup for new include directory locations.
 *
 *  Revision 1.15  1997/09/06 04:05:46  taylorj
 *  Major: New maintainer (me).  Also reformatted all VGA driver source to GGI
 *  specs and  converted all debugging prints to use the newer debugging macros.
 *
 *  Revision 1.14  1997/03/06 17:51:42  ezrec
 *  Major: Splitline support from Massimiliano Ghila (and the associated
 *  evil that the font_pointer needed)
 *
 *  Revision 1.13  1997/02/28 07:49:10  ezrec
 *  Minor: in vga_text16.inc, fixed block-cursor w/respect to set_split.
 *  in vga_text16_enh.inc, noticed font-cursor bug w/respect to
 *  set_split, so disabled set_split (fun with #ifdef) for now...
 *
 *  Revision 1.12  1997/02/28 07:35:51  ezrec
 *  Minor: Block cursor now (should) work properly with splitline...
 *
 *  Revision 1.11  1997/02/28 06:59:56  ezrec
 *  Minor: set_splitline functionality is back in - and thanks to jochen,
 *  working!
 *
 *  Revision 1.10  1997/02/26 05:28:41  ezrec
 *  Minor: Thanks to Massimiliano Ghilardi <ghilardi@cibs.sns.it>, the
 *         font-cursor works very well... SLOW_POINTER is defined,
 *         as I need it, but try without SLOW_POINTER and mail me or
 *         M.G. the results...
 *
 *  Revision 1.9  1997/02/21 13:57:55  ezrec
 *  Minor: Added 'SLOW_POINTER' option, for people who don't like waiting
 *  	for vertical retrace... Default is undefined.
 *
 *  Revision 1.8  1997/02/20 21:59:08  ezrec
 *  Minor: jmcc fixes: keypress_wait bug on 2.1.x (console.c)
 *         Jochen Hoenicke <Jochen.Hoenicke@Informatik.Uni-Oldenburg.DE> fixed:
 *  	  IBM/vga drivers
 *  		No-display-on-driver-load bug (woo woo!)
 *  		set_origin/Attribute handling bugs...
 *  		Added TEXT_A000_64, for those of you with MDA or
 *  			Hercules cards (it's now the default, anyway)
 *
 *  Revision 1.7  1997/02/15 23:19:45  ezrec
 *  Minor: Massimiliano Ghilardi's [ghilardi@cibs.sns.it] XOR cursor patches...
 *
 *  Revision 1.6  1997/02/12 06:25:00  ezrec
 *  Minor: Font-cursor is much smoother now, as is font-setting...
 *
 *  Revision 1.5  1997/02/09 18:40:21  ezrec
 *  Minor: Pointer is now an 'XOR' cursor, and doesn't change the attribute
 *  	or the cell it's on. Moved the 'charcell' type to
 *  	$GGIHOME/include/IBM/vga_text16.h
 *
 *  Revision 1.4  1997/02/09 05:15:23  ezrec
 *  Major: Added support for cursors up to 16x16 bits in size, clipping
 *         on the edge of the text frame, and the getpointer utility
 *         for automatically generating pointer.h files from X11 cursor
 *         masks
 *
 *  Revision 1.3  1997/02/08 23:14:05  ezrec
 *  Minor: Fixed 'thinko' that led to Oopses under 2.1.25, and now I do
 *  pointer hiding/unhiding more intelligently..  (Tin was having a
 *  HELLUVA time w/the pointer before, but it's _real_ nice now...)
 *
 *  Revision 1.2  1997/02/08 19:39:07  ezrec
 *  Minor: Cursor now uses as few charcells as possible, and is better shaped..
 *
 *  Revision 1.1  1997/02/08 08:58:56  ezrec
 *  Major: Set block cursor as default, and added font-hack pointer.  Modify
 *        pointer.h to suit your taste.... Still 'wiggles' a lot, but better
 *        than it used to be...
 *
 *  Revision 1.4  1997/02/06 03:33:37  ezrec
 *  Minor: Moahaha! set_origin now works properly, so we have _really_
 *  	fast A0000-based text modes (128K of mind-numbing text
 *      scrolling power!) Undef 'TEXT_A000' for B8000 (32k) text
 *  	modes, and Undef 'CURSOR_HW' to get a solid cursor.
 *  	(for laptops...)
 *
 *  Revision 1.3  1997/02/05 22:55:51  ezrec
 *  Major:  Fixed the 'cursor follows' point problem, and hardware/software
 *  	cursor is selectable via the 'CURSOR_HW' define in vga.c
 *  	NOTE: You _must_ change video modes before anything will display..
 *
 *  Revision 1.2  1997/02/04 04:33:46  ezrec
 *  Minor: Small bugfixes to cursor and pointer code, but not fixed yet....
 *
 *  Revision 1.1  1997/02/03 15:06:37  ezrec
 *  Major: IBM vga driver (initial version, taken from GGI 0.0.8)
 *
 * ---------------------------------------------------------------------------
 *
 *  This code was derived from the following source of information:
 *  [1]     Xfree 86 vga driver
 *  [2]     Zen of Graphics Programming (Michael Abrash)
 */

#ifndef _VGA_TEXT16_INC_
#define _VGA_TEXT16_INC_

#undef  DEBUG_P		/* Debug Font-pointer (_very_ verbose!) */

/* GGI includes */

#include "IBM/vga_text16.h"
#include "pointer_mask.h"

#define CURSOR_MASK	0xE8	/* Attribute mask (bg,fg) */
#define POINTER_MASK	0xFF	/* Attribute mask (bg,fg) */

#define PCELLSIDEX	3	/* Horizontal length of pointer cell box */
#define PCELLSIDEY	3	/* Vertical length of pointer cell box */

#define PCELLTOTAL	(PCELLSIDEX * PCELLSIDEY)		
#define PCELLOF(fp, xsize, i) ((fp) + (((i) / PCELLSIDEX) * (xsize)) + ((i) % PCELLSIDEX))

typedef struct text16_point_s
{
	int	visible;	/* Should I visible the old data? */
	int	offset;		/* Offset of old point (charcells) */
	int	offset2;	/* Extra offset, due to split line */
	int	splity;         /* Split line in old point (0..PCELLSIDEY) */
	ggi_uint	maxx;		/* Max visible x of old point (charcells) */
	ggi_uint	maxy;		/* Max visible y of old point (charcells) */
	int	bitoffset;	/* Offset of old point (dots) */
	charcell data[PCELLTOTAL];	/* Data at old point */
} text16_point;
	
struct text16_data_s
{
	ggi_uint		magic;
	int		font_changed;
	unsigned long	origin;
	int		splitline;
	text16_point  	pointer;
#ifndef CURSOR_HW
	text16_point	cursor;
#endif
} text16_data[MAX_NR_DISPLAYS];

#define ORIGIN 		(data->origin)
#define SPLITLINE	(data->splitline)
#define FONTCHANGED	(data->font_changed)

#define PVISIBLE	(data->pointer.visible)
#define POFFSET1	(data->pointer.offset)
#define POFFSET2	(data->pointer.offset2)
#define PSPLITY		(data->pointer.splity)
#define PMAXX		(data->pointer.maxx)
#define PMAXY		(data->pointer.maxy)
#define PBITOFFSET	(data->pointer.bitoffset)
#define PDATA		(data->pointer.data)

#ifndef CURSOR_HW
#define CVISIBLE	(data->cursor.visible)
#define COFFSET		(data->cursor.offset)
#define CDATA		(data->cursor.data[0])
#endif

#define GET_TEXT_DATA struct text16_data_s *data = \
		(struct text16_data_s *)KGIM_PRIVATE(dpy, driver)

static void text16_data_init(void)
{
	memset(text16_data, 0, sizeof(struct text16_data *) * MAX_NR_DISPLAYS);
}
	
static struct text16_data_s *text16_data_realloc(struct text16_data_s *old)
{
	int i;

	if (old == NULL || old->magic != TEXT16_MAGIC) {
		for (i = 0; i < MAX_NR_DISPLAYS; i++) 
			if (text16_data[i].magic != TEXT16_MAGIC) {
				memset(&(text16_data[i]), 0,
				 sizeof(struct text16_data_s));
				text16_data[i].magic = TEXT16_MAGIC;
				return &(text16_data[i]);
			}
		return NULL;
	}
	memset(old, 0, sizeof(struct text16_data_s));
	old->magic = TEXT16_MAGIC;
	return old;
}

/* Text-font based mouse pointer code */

/* Ascii codes of font locs. to abuse for our purposes */
static unsigned char pointer_char[PCELLTOTAL] = {
	0xC5, 0xC6, 0xC7,
	0xD6, 0xD7, 0xD8,
	0xCF, 0xDD, 0xDE
};

#if DEBUG_LEVEL > 0
static void print_mask(const char *msg, ggi_uint *mask, ggi_uint texty)
{
	int i, j;
	char buff[33];

	printk("%s\n", msg);

	buff[32] = 0;
	for (i = 0; i < texty * 2; i++) {
		for (j = 0; j < 32; j++) 
			buff[31 - j] = (mask[i] & (1 << j)) ? 'O' : ' ';    
		printk("  [%.2d]%s\n", i, buff);
	}
}
#endif /* DEBUG_LEVEL > 0 */

typedef struct text16_font_reg_s 
{
	uchar SR2,SR4;
	uchar GR4,GR5,GR6;
} text16_font_reg;

#define FONT_GET 0
#define FONT_SET 1

#ifdef NO_LONGER_USED_CODE
static void text16_font_access(int getset, int chr, uchar *charcell)
{
	text16_font_reg regs;
	uchar *fontbuff = (uchar *)(phys_to_virt(0xA0000)) + (chr << 5);
	IS_CRITICAL;

#ifdef SLOW_POINTER
	WAIT_VRETRACE();
#endif
	ENTER_CRITICAL;
	regs.SR2 = inbSEQ(0x02);
	regs.SR4 = inbSEQ(0x04);	/* Save register state        */
	regs.GR4 = inbGRC(0x04);
	regs.GR5 = inbGRC(0x05);
	regs.GR6 = inbGRC(0x06);

    outbSEQ(SR02_PLANE2, 0x02);		/* Font data goes to plane 2  */
    outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

    outbGRC(2, 0x04);			/* Read plane 2               */
    outbGRC(GR5_WRITEMODE0, 0x05);
    outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);

    if (getset == FONT_GET) 
		memcpy(charcell, fontbuff, 32);
    else
		memcpy(fontbuff, charcell, 32);

	outbSEQ(regs.SR2, 0x02);
	outbSEQ(regs.SR4, 0x04);	/* Restore registers  */
	outbGRC(regs.GR4, 0x04);
	outbGRC(regs.GR5, 0x05);
	outbGRC(regs.GR6, 0x06);
	LEAVE_CRITICAL;
}
#endif /* NO_LONGER_USED_CODE */

static inline void text16_font_enter(text16_font_reg *regs) 
{
	IS_CRITICAL;

#ifdef SLOW_POINTER
	WAIT_VRETRACE();
#endif
	ENTER_CRITICAL;
	regs->SR2 = inbSEQ(0x02);
	regs->SR4 = inbSEQ(0x04);	/* Save register state        */
	regs->GR4 = inbGRC(0x04);
	regs->GR5 = inbGRC(0x05);
	regs->GR6 = inbGRC(0x06);

	outbSEQ(SR02_PLANE2, 0x02);	/* Font data goes to plane 2  */
	outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

	outbGRC(2, 0x04);		/* Read plane 2               */
	outbGRC(GR5_WRITEMODE0, 0x05);
	outbGRC(GR6_MAP_A0_64, 0x06);
	LEAVE_CRITICAL;
}

static void inline text16_font_leave(text16_font_reg *regs)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outbSEQ(regs->SR2, 0x02);
	outbSEQ(regs->SR4, 0x04);	/* Restore registers  */
	outbGRC(regs->GR4, 0x04);
	outbGRC(regs->GR5, 0x05);
	outbGRC(regs->GR6, 0x06);
	LEAVE_CRITICAL;
}

static inline void direct_text16_font_read(int chr, uchar *charcell)
{
	memcpy(charcell, (uchar *)(phys_to_virt(0xA0000)) + (chr << 5), 32);
}

static inline void direct_text16_font_write(int chr, uchar *charcell)
{
	memcpy((uchar *)(phys_to_virt(0xA0000)) + (chr << 5), charcell, 32);
}

static void place_pointer(charcell *fb, ggi_uint xsize, ggi_uint dx, ggi_uint dy,
			  ggi_uint texty, int isdelt, struct text16_data_s *data)
{
	int i, j;
	ggi_uint mask[48];
    uchar font[PCELLTOTAL * 32], *cp;
	text16_font_reg regs;
	charcell *fb1 = fb + POFFSET1, *fb2 = fb + POFFSET2;
	charcell *cell = PDATA;
	ggi_uint maxx = PMAXX, maxy = PMAXY;
	int splity = PSPLITY;
    
	if (dy > 16) {
		NOTICE1("PtrHack: Can't deal with characters >16 bits high");
		return;
	}
   
#define LOOPOVERxyi { register int x,y; for (y=0;y<maxy;y++) { \
			i = y * PCELLSIDEX; \
			fb = (y < splity) ? fb1 : fb2; \
			for (x = 0; x < maxx; x++, i++) {
#define ENDLOOPxyi  }}}

    /* Has anything changed since the last time we mucked with the fonts? */ 
	if (!isdelt) {
		LOOPOVERxyi
			isdelt |= (cell[i] != *PCELLOF(fb,xsize,i));
		ENDLOOPxyi
		    
			/* Nope - same characters as we used last time.  Why redo our effort? */
			
			if (!isdelt) {
				LOOPOVERxyi
					CHARSET(PCELLOF(fb, xsize, i), pointer_char[i]);
				ENDLOOPxyi
					return;
			}
	}

    /* Get the characters cells and their associated fonts */
    LOOPOVERxyi
	cell[i] = *PCELLOF(fb, xsize, i);	
    ENDLOOPxyi

    text16_font_enter(&regs);

    LOOPOVERxyi
	direct_text16_font_read(CHAR(cell[i]), font + (i << 5));
    ENDLOOPxyi

    /* Copy the old chars, as a bitmap, into mask[]  */
    memset(mask, 0, sizeof(mask)); 
    LOOPOVERxyi
	cp = font + (i << 5);
	for (j = 0;j < texty; j++) 
		mask[j + y * texty] |= (*(cp++)) << ((3 - x) * 8);
    ENDLOOPxyi

#if DEBUG_LEVEL > 0
    print_mask("Original", mask, texty);
#endif

    /* Mask with the And and Or masks */
    for (i = 0; i < POINTER_MASK_LEN; i++) {
	mask[dy + i] |= (pointer_or_mask[i] >> dx);
	mask[dy + i] ^= (pointer_xor_mask[i] >> dx);
    }
         
#if DEBUG_LEVEL > 0
    print_mask("W/Pointer", mask, texty);
#endif

    /* Now write the new data */
    LOOPOVERxyi
	cp = font + (i << 5);
	for (j = 0; j < texty; j++) 
		*(cp++) = (mask[j + y * texty] >> ((3 - x) * 8)) & 0xFF;
    ENDLOOPxyi

    LOOPOVERxyi
	direct_text16_font_write(pointer_char[i], font + (i << 5));
    ENDLOOPxyi

    text16_font_leave(&regs);

    LOOPOVERxyi
	CHARSET(PCELLOF(fb, xsize, i), pointer_char[i]);
    ENDLOOPxyi
}

static void erase_pointer(charcell *fb, ggi_uint xsize, struct text16_data_s *data)
{
	int i;
	charcell *fb1 = fb + POFFSET1, *fb2 = fb + POFFSET2;
	ggi_uint maxx = PMAXX, maxy = PMAXY;
	int splity = PSPLITY;

	LOOPOVERxyi
		*PCELLOF(fb, xsize, i) = PDATA[i];
	ENDLOOPxyi
}

#undef LOOPOVERxyi
#undef ENDLOOPxyi

/* Font setting */

static void text16_do_set_font(int no, int from, int to, struct kgi_font *font)
{
	text16_font_reg fontregs;
	int i;
	char *fbuf, *data;
	register int cnt = 32 - font->dy;

	if (font == NULL)
		return;

	fbuf = (char *) phys_to_virt(0x000A0000);
	fbuf += (no * 0x2000) + (from << 5);
	data = font->data + font->dy * from;
	if (!data)
		return;

	text16_font_enter(&fontregs);
	/* NOTE: no printk()'s are allowed in here unless klogd is active! */

	for (i = from; i <= to; i++) {
		memmove(fbuf, data, font->dy);
		memset(fbuf + font->dy, 0, cnt);
		fbuf += 32;
		data += font->dy;
	};

	text16_font_leave(&fontregs);

#ifdef CURSOR_HW
	/* Set up cursor */
	outbCRT((inbCRT(0x0A) & ~CR0A_CURSOR_START_MASK)
		| ((font->base+1) & CR0A_CURSOR_START_MASK), 0x0A);
	outbCRT((inbCRT(0x0B) & ~CR0B_CURSOR_END_MASK)
		| ((font->base+2) & CR0B_CURSOR_END_MASK), 0x0B);
#endif
}

static void text16_set_font(struct kgi_display *dpy, struct kgi_font *font,
			    ggi_uint from, ggi_uint to)
{
	GET_TEXT_DATA;
	text16_do_set_font(0, from, to, font);

	FONTCHANGED = 1;
}

static void text16_set_origin(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_TEXT_DATA;
	IS_CRITICAL;
	static ggi_uint last_xdot = -1;
	static ggi_uint last_ydot = -1;
	register ggi_uint xpix = x / dpy->mode.request.dpp.x;
	register ggi_uint xdot = x % dpy->mode.request.dpp.x;
	register ggi_uint ypix = y / dpy->mode.request.dpp.y;
	register ggi_uint ydot = y % dpy->mode.request.dpp.y;

	register ggi_uint offset = xpix + ypix*dpy->fb.size.x;

	ENTER_CRITICAL;

		if (ORIGIN != offset) {

			outbCRT(offset >> 8, CRT_STARTADDR_H);
			outbCRT(offset, CRT_STARTADDR_L);
			ORIGIN = offset;
		}

		if (last_ydot != ydot) {

			outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
			last_ydot = ydot;
		}

		if (last_xdot != xdot) {

			register ggi_uint hpan;

			if (dpy->mode.request.dpp.x == 9) {

				hpan = (xdot == 0)
					? 0x08 : (xdot-1) & AR13_DATA_MASK;
			} else {

				hpan = xdot & AR13_DATA_MASK;
			}
			WAIT_HVRETRACE();
			outbATC(hpan, ATC_HPAN);
			outb(ATCI_ENABLEDISPLAY, ATCI);
			last_xdot = xdot;
		}

	LEAVE_CRITICAL;
}

static void text16_set_split(struct kgi_display *dpy, int line)
{
	GET_TEXT_DATA;
	register uchar foo;

	outbCRT(line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? inbCRT(0x07) | CR07_LINECOMPARE_B8
		: inbCRT(0x07) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	foo = (line & 0x0200)
		? inbCRT(0x09) | CR09_LINECOMPARE_B9
		: inbCRT(0x09) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);

	SPLITLINE = line;
}

#define	text16_set_frame	NULL	/* !!! implement! */

#define DPPX            (dpy->mode.request.dpp.x)
#define DPPY            (dpy->mode.request.dpp.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

void text16_show_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_TEXT_DATA;
	ulong newoff1 = ORIGIN + (x / DPPX) + (y / DPPY) * VIRTUALX;
	ulong newoff2 = 0;
        register charcell *p = (void *) dpy->fb.pbuf;
	ggi_uint bitoff = x + y * VISIBLEX;
	ggi_uint dx, dy, clipx, clipy, maxx, maxy;
	int splity = PCELLSIDEY - 1;
	
	if (SPLITLINE && SPLITLINE < VIRTUALY * DPPY &&
	    y+POINTER_MASK_LEN > SPLITLINE) {
		newoff2 = (x / DPPX) + ((y - SPLITLINE) / DPPY) * VIRTUALX;
		if (y >= SPLITLINE) {
			newoff1 = newoff2;
			newoff2 = 0;
		}
		else 
			splity = SPLITLINE / DPPY - y / DPPY + 1;
	}

	DEBUG1("PShow(%c):POFFSET1=0x%.8x POFFSET2=0x%.8x PBOFFSET=0x%.8x x=%d,y=%d\n",
		PVISIBLE ? 'V' : 'i', POFFSET1, POFFSET2, PBITOFFSET, x, y);

 	ASSERT((x < VISIBLEX * DPPX) && (y < VISIBLEY * DPPY));

	if (PVISIBLE) 
		erase_pointer(p, VIRTUALX, data);

	dx = x % DPPX;
	dy = y % DPPY;
	clipx = VIRTUALX - (x / DPPX);
	clipy = VIRTUALY - (y / DPPY);

	maxx = ((POINTER_MASK_WIDTH - 1 + dx) / 8) + 1;
	maxx = (maxx > clipx) ? clipx : maxx;
	maxy = ((POINTER_MASK_LEN - 1 + dy) / DPPY) + 1;
	maxy = (maxy > clipy) ? clipy : maxy;

	if (splity < 0)
		splity = 0;
	else if (splity > maxy)
		splity = maxy;

        POFFSET1 = newoff1;
	POFFSET2 = newoff2;
	PSPLITY = splity;
	PMAXX = maxx;
	PMAXY = maxy;
	
	place_pointer(p, VIRTUALX, dx, dy, DPPY, 
		      bitoff != PBITOFFSET || FONTCHANGED, data);

	PBITOFFSET = bitoff;
	PVISIBLE = 1;
	FONTCHANGED = 0;
}

static void text16_hide_pointer(struct kgi_display *dpy)
{
	GET_TEXT_DATA;
	register charcell *p = (void *) dpy->fb.pbuf;

	DEBUG1("PHide(%c): POFFSET1=0x%.8x POFFSET2=0x%.8x PBOFFSET=0x%.8x\n",
		PVISIBLE ? 'V' : 'i', POFFSET1, POFFSET2, PBITOFFSET);

	if (PVISIBLE) {
		erase_pointer(p, VIRTUALX, data);
		PVISIBLE = 0;
	}
}

#define text16_undo_pointer     text16_hide_pointer

#ifdef CURSOR_HW
static void text16_show_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_TEXT_DATA;
	register ulong newoff;
 	IS_CRITICAL;

        ASSERT((x < VISIBLEX * DPPX) && (y < VISIBLEY * DPPY));

	if (SPLITLINE && y >= SPLITLINE) 
		newoff = (x / DPPX) + ((y - SPLITLINE) / DPPY) * VIRTUALX;
	else
		newoff = ORIGIN + (x / DPPX) + (y / DPPY) * VIRTUALX;

	ENTER_CRITICAL;
	outbCRT(newoff >> 8, CRT_CURSORADDR_H);
	outbCRT(newoff & 0xFF, CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outbCRT(0xFF, CRT_CURSORADDR_H);
	outbCRT(0xFF, CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

/* This is badly wrong. Hardware cursor doesn't need and undo function! We 
 * should put NULL here and update kgi.c to actually check if dpy->undo_xxxx
 * are not NULL before calling them.
 */
#define text16_undo_cursor      text16_hide_cursor
#else
static void text16_show_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_TEXT_DATA;
	register ulong newoff;
	register charcell *p = (void *) dpy->fb.pbuf;
	register charcell cell;

	ASSERT((x < VISIBLEX * DPPX) && (y < VISIBLEY * DPPY));

	if (SPLITLINE && y >= SPLITLINE) 
		newoff = (x / DPPX) + ((y - SPLITLINE) / DPPY) * VIRTUALX;
	else
		newoff = ORIGIN + (x / DPPX) + (y / DPPY) * VIRTUALX;

	if (!CVISIBLE || newoff != COFFSET) {
		cell = *(p + newoff);
		ATTRSET(p + newoff, ATTR(cell) ^ CURSOR_MASK);
	} else 
		return;

	if (CVISIBLE)
		*(p + COFFSET) = CDATA;
	else
		CVISIBLE = 1;

	COFFSET = newoff;
	CDATA = cell;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	GET_TEXT_DATA;
	register ushort *p = (void *) dpy->fb.pbuf;

	if (CVISIBLE)
		*(p + COFFSET) = CDATA;
	
	CVISIBLE = 0;
}

#define text16_undo_cursor      text16_hide_cursor
#endif /* CURSOR_HW */

#undef SPLITLINE
#undef ORIGIN

#undef PVISIBLE
#undef POFFSET1
#undef POFFSET2
#undef PSPLITY
#undef PMAXX
#undef PMAXY
#undef PDATA

#ifndef CURSOR_HW
#undef CVISIBLE
#undef COFFSET
#undef CDATA
#endif

#endif /* _VGA_TEXT16_INC */

