/* --------------------------------------------------------------------------
**	text16 operators
** --------------------------------------------------------------------------
**
**	Copyright (C)	1997		Jonas Borgstrom
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	This code was derived from the following source of information:
**	[1]     Xfree 86 vga driver
**	[2]     Zen of Graphics Programming (Michael Abrash)
*/

#include <ggi/module.h>

#include <Matrox/mga_common_regs.h>
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga_common_text16.h>
#include <Matrox/mga_common.h>
/* Driver specific things are included by mga_common_*.h */

#define TEXTX           (dpy->mode.request.dpp.x)
#define TEXTY           (dpy->mode.request.dpp.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

static void do_set_font(struct kgi_display *dpy, int no, ggi_uint first,
			ggi_uint last, struct kgi_font *font)
{
	IS_CRITICAL;
	GET_PRIVATE;
	uint8 *data;
	mgafont *fbuf = (mgafont *)MGA_FRAMEBUFFER.baseptr;
	int i,j;
#ifdef DRIVER_mga2064 /* Causes flickering on the Mystique */
        uint8 tmp;
#endif   
	/*
	 * Enable framebuffer
	 */
#ifdef DRIVER_mga2064 /* Causes flickering on the Mystique */
	mem_out8(CRTX_EXT3, MMIO_CRTXI);
        tmp = mem_in8(MMIO_CRTXD);
        mem_out8(tmp | CRTX_MGAMODE, MMIO_CRTXD);
#endif
	if (font == NULL) {
		return;
	}
	fbuf += (no * 256 + first) * 32;
	data = font->data + font->dy * first;

	/*	Display off while setting font.
	 */
	ENTER_CRITICAL;
	SEQ_out8(dpy, SEQ_in8(dpy, SEQ_CLOCK) | SR01_DISPLAY_OFF, SEQ_CLOCK);
	LEAVE_CRITICAL;

	/* NOTE	no printk()'s are allowed in here unless klogd is active !
	*/

	if (data) {
		/*	set font if data is valid
		*/
		int cnt = 32 - font->dy;

		for (i = first; i <= last; i++) {
			for (j = font->dy; j--; ) {
				mem_out8(*(data++), &((fbuf++)->font));
			}
			for (j = cnt; j--; ) {
				mem_out8(0, &((fbuf++)->font));
			}
		}
	} else {
		/*	clear font
		*/
		for (i = first*32; i <= last*32; i++ ) {
			mem_out8(0, &((fbuf++)->font));
		}
	}

	/*	restore hardware state
	*/
	ENTER_CRITICAL;

	CRT_out8(dpy, (CRT_in8(dpy, CRT_CURSORSTART) & 
		       ~CR0A_CURSOR_START_MASK) | 
		 ((font->base + 1) & CR0A_CURSOR_START_MASK), CRT_CURSORSTART);
	CRT_out8(dpy, (CRT_in8(dpy, CRT_CURSOREND) & 
		       ~CR0B_CURSOR_END_MASK) |
		 ((font->base + 2) & CR0B_CURSOR_END_MASK), CRT_CURSOREND);
	SEQ_out8(dpy, SEQ_in8(dpy, SEQ_CLOCK) & ~SR01_DISPLAY_OFF, SEQ_CLOCK);
#ifdef DRIVER_mga2064 /* Causes flickering on the Mystique ??? */
	mem_out8(CRTX_EXT3, MMIO_CRTXI);
        mem_out8(tmp, MMIO_CRTXD);
#endif
	LEAVE_CRITICAL;
}

void mga_text_set_font(struct kgi_display *dpy, struct kgi_font *font,
		       ggi_uint from, ggi_uint to)
{
	do_set_font(dpy, 0, from, to, font);
	do_set_font(dpy, 1, from, to, font);
}

void mga_text_set_origin(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	IS_CRITICAL;
	GET_PRIVATE;
	ggi_uint xpix = x / TEXTX;
	ggi_uint xdot = x % TEXTX;
	ggi_uint ypix = y / TEXTY;
	ggi_uint ydot = y % TEXTY;

	register ggi_uint offset = xpix + ypix*dpy->fb.size.x;
	private->cp.origin = private->cp.frame + offset;

	ENTER_CRITICAL;

	if (private->cp.org_offs != offset) {
		CRT_out8(dpy, offset >> 8, CRT_STARTADDR_H);
		CRT_out8(dpy, offset,      CRT_STARTADDR_L);
		private->cp.org_offs = offset;
	}

	if (private->cp.org_ydot != ydot) {
		CRT_out8(dpy, ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
		private->cp.org_ydot = ydot;
	}

	if (private->cp.org_xdot != xdot) {
		if (dpy->mode.request.dpp.x == 9) {
			ATC_out8(dpy, (xdot == 9) ? 0x08 :
				 (xdot-1) & AR13_DATA_MASK, ATC_HPAN);
		} else {
			ATC_out8(dpy, xdot & AR13_DATA_MASK, ATC_HPAN);
		}
		private->cp.org_xdot = xdot;
	}

	LEAVE_CRITICAL;
}

/* In MGA gfx modes, the splitline doesn't work.  Obviously it is only
** there to allow the vertical line interrupt.
*/
void mga_set_split(struct kgi_display *dpy, int line)
{
	IS_CRITICAL;
	GET_PRIVATE;
	uint8 foo;

        line *= dpy->mode.magnify.y;
        if (private->cp.split == line)
		return;	
	
	private->cp.split = line;

	ENTER_CRITICAL;
	CRT_out8(dpy, line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? CRT_in8(dpy, 0x07) |  CR07_LINECOMPARE_B8
		: CRT_in8(dpy, 0x07) & ~CR07_LINECOMPARE_B8;
	CRT_out8(dpy, foo, 0x07);

	foo = (line & 0x0200)
		? CRT_in8(dpy, 0x09) |  CR09_LINECOMPARE_B9
		: CRT_in8(dpy, 0x09) & ~CR09_LINECOMPARE_B9;
	CRT_out8(dpy, foo, 0x09);

	/* Set splitline in E-CRTC registers.  Only used in MGA modes
	   but if it's VGA, it's zero, so setting it never hurts. */
	ECRT_out8(dpy, (line & 0x0400)
		  ? ECRT_in8(dpy, 2) | 0x80 : ECRT_in8(dpy, 2) & ~0x80, 2);

	LEAVE_CRITICAL;

	DEBUG2("Set split: %d\n", line);
}

/*
**	Cursor handling
*/

/*	hardware cursor
*/
void mga_show_hw_text_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	IS_CRITICAL;
	GET_PRIVATE;
	ggi_uint pos = private->cp.origin + 
	               (x / TEXTX) + (y / TEXTY)*dpy->fb.size.x;

	pos = (pos < private->cp.fb_size) ? pos : pos - private->cp.fb_size;
	ASSERT(pos < private->cp.fb_size);

	DEBUG2("cursor (%i,%i) = %i", x, y, pos);
	ENTER_CRITICAL;
	CRT_out8(dpy, pos >> 8, CRT_CURSORADDR_H);
	CRT_out8(dpy, pos,      CRT_CURSORADDR_L);
	CRT_out8(dpy, CRT_in8(dpy, CRT_CURSORSTART) & ~CR0A_CURSOR_OFF,
		 CRT_CURSORSTART);
	LEAVE_CRITICAL;
}

void mga_hide_hw_text_cursor(struct kgi_display *dpy)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	CRT_out8(dpy, CRT_in8(dpy, CRT_CURSORSTART) | CR0A_CURSOR_OFF,
		 CRT_CURSORSTART);
	LEAVE_CRITICAL;
}

/*	software pointer and cursor
*/
#define	PTRAND	(private->cp.ptr_and)
#define	PTRXOR	(private->cp.ptr_xor)
#define	PTROLD	(private->cp.ptr_old)

#define	CURAND	(private->cp.cur_and)
#define	CURXOR	(private->cp.cur_xor)
#define	CUROLD	(private->cp.cur_old)

void mga_show_sw_text_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_PRIVATE;
	uint32 old = PTROLD;
	uint32 cur = CUROLD;
	uint32 new = private->cp.origin +
		     (x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	mgatext *p = (mgatext *) dpy->fb.pbuf;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	new = (new < private->cp.fb_size) ? new : new - private->cp.fb_size;
	ASSERT(new < private->cp.fb_size);

	if (old & 0x8000) mem_out16(old >> 16, p + (old & 0x7FFF));

	p += new;

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	PTROLD = new | 0x8000;
	mem_out16((mem_in16(p) & PTRAND) ^ PTRXOR, p);
}

void mga_hide_sw_text_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;
	uint32 old = PTROLD;
	mgatext *p = (mgatext *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) mem_out16(old >> 16, p + (old & 0x7FFF));

	PTROLD = 0;
}

void mga_show_sw_text_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_PRIVATE;
	uint32 old = CUROLD;
	uint32 cur = PTROLD;
	uint32 new = private->cp.origin +
		     (x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	mgatext *p = (mgatext *) dpy->fb.pbuf;

	ASSERT(((x / TEXTX) < VISIBLEX) && ((y / TEXTY) < VISIBLEY));

	new = (new < private->cp.fb_size) ? new : new - private->cp.fb_size;
	ASSERT(new < private->cp.fb_size);

	if (old & 0x8000) mem_out16(old >> 16, p + (old & 0x7FFF));

	p += new;

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	CUROLD = new | 0x8000;
	mem_out16((mem_in16(p) & CURAND) ^ CURXOR, p);
}

void mga_hide_sw_text_cursor(struct kgi_display *dpy)
{
	GET_PRIVATE;
	uint32 old = CUROLD;
	mgatext *p = (mgatext *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) mem_out16(old >> 16, p + (old & 0x7FFF));
	CUROLD = 0;
}

#undef	PTRAND
#undef	PTRXOR
#undef	PTROLD

#undef	CURAND
#undef	CURXOR
#undef	CUROLD

#undef	TEXTX
#undef	TEXTY
#undef	VIRTUALX
#undef	VISIBLEX
#undef	VISIBLEY

/* ----------------------------------------------------------------------------
**	Millenium and Mystique text framebuffer operation
** ----------------------------------------------------------------------------
**
** Why is there a special version for text functions for the Matrox cards?
** Because in order to do multihead, we have to turn off all response to 
** VGA resources, including VGA framebuffer.  But we still want to have
** textmodes, so there's only the "real" framebuffer left, which is controlled
** by PCI resources.  Unfornunately (actually I like it more), the layout of
** textmode in the MGA framebuffer is different from the VGA one, so we have 
** to provide specialized text operations.
**
** ----------------------------------------------------------------------------
*/

#define	FBSIZE	(fb->size.x*fb->size.y)

static void mgatext_move_io(kgi_pixmap *fb, ggi_uint dst, ggi_uint src, 
			    ggi_uint cnt)
{
	mgatext *pbuf = ((mgatext *)fb->pbuf);
	mgatext *pdst;
	mgatext *psrc;

	ASSERT((src + cnt) <= FBSIZE);
	ASSERT((dst + cnt) <= FBSIZE);

	psrc = pbuf + src;
	pdst = pbuf + dst;
	if (dst < src) {
		while (cnt--) 
		        (pdst++)->ca = (psrc++)->ca;
	} else {
		psrc += cnt;
		pdst += cnt;
		while (cnt--)
		        (--pdst)->ca = (--psrc)->ca;
	}
}

static void mgatext_set_io(kgi_pixmap *fb, ggi_uint src, ggi_pixel val, 
			   ggi_uint cnt)
{
	mgatext *pbuf = ((mgatext *) fb->pbuf) + src;

	ASSERT((src + cnt) <= FBSIZE);

	while (cnt--)
	        (pbuf++)->ca = val;
}

static void mgatext_invert_io(kgi_pixmap *fb, ggi_uint start, ggi_uint end)
{
	mgatext *pix = ((mgatext *) fb->pbuf) + start;

	ASSERT((start < FBSIZE) && (end < FBSIZE));
	ASSERT(start <= end);

	while (start++ <= end) {
		(pix++)->ca = (pix->ca & 0x88FF) |
		  ((pix->ca<<4) & 0x7000) | ((pix->ca>>4) & 0x0700);
	}
}

static void mgatext_update_io(kgi_pixmap *fb, ggi_uint pos, ggi_pixel val)
{
	ASSERT(pos < FBSIZE);

	(((mgatext*) fb->pbuf) + pos)->ca = val;
}

static void mgatext_flush_io(kgi_pixmap *fb)
{
}

static void mgatext_copyto_io(kgi_pixmap *fb, void *buf, ggi_uint start, 
			      ggi_uint cnt)
{
	mgatext *dst = ((mgatext *)fb->pbuf) + start;
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--)
	        (dst++)->ca = *((ushort *)buf)++;
}

static void mgatext_copyfrom_io(kgi_pixmap *fb, void *buf, ggi_uint start, 
				ggi_uint cnt)
{
	mgatext *src = ((mgatext *)fb->pbuf) + start;
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--)
	        *((ushort *)buf)++ = (src++)->ca;
}

/*	We have operations for I/O only.  When a console is unmapped,
**      the characters are stored "normal".
*/
kgi_fb_ops mgatext_color_io = {
	mgatext_move_io,
	mgatext_set_io,
	mgatext_invert_io,
	mgatext_update_io,
	mgatext_flush_io,
	mgatext_copyfrom_io,
	mgatext_copyto_io,
	NULL,
	NULL,
	NULL,
	NULL
};
