/* ---------------------------------------------------------------------------
 *  Tseng Labs ET6000 fb operators
 * --------------------------------------------------------------------------
 *
 *  Copyright (C) 1997, The Linux-GGI Project
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 *
 *  $Log: et6000_fb.inc,v $
 *  Revision 1.2  1998/02/21 18:42:04  ursetto
 *  text->dpp changeover
 *
 *  Revision 1.1  1998/01/08 07:45:59  ursetto
 *  Initial check-in.
 *
 *
 * ---------------------------------------------------------------------------
 *
 *  This code was derived from the following sources of information:
 *  [1]     IBM VGA driver
 *  [2]     Prelease VGADOC5 by Finn Thoegersen
  */

#ifndef _ET6000_FB_INC
#define _ET6000_FB_INC

/* Smooth-scrolling 24 bit probably doesn't work. */
static void et6000_fb_set_origin(struct kgi_display *dpy, uint x, uint y)
{
	IS_CRITICAL;
	static uint last_xdot = -1;
	static uint last_ydot = -1;
	static uint old_offset = -1;
	register uint xpix=0,xdot=0;
	register uint ypix=0,ydot=0;
	register uint offset;
	uint fb_width=0;

	/* The chipset_command fxns have already made sure we're in bounds
	 * and have accounted for splitline. */

	switch (dpy->mode.request.graphtype) {

/* text16_set_origin handles this functionality for text mode. */
/* May have to reinsert this, though. */
/*      case GT_TEXT16:
		xpix = x / dpy->mode.request.dpp.x;
		xdot = x % dpy->mode.request.dpp.x;
		ypix = y / dpy->mode.request.dpp.y;
		ydot = y % dpy->mode.request.dpp.y;
		fb_width = dpy->mode.request.virt.x * 2;
*/			
			
		case GT_1BIT:
		case GT_4BIT:
			xpix = x / 8;
			xdot = x % 8;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x / 8;
			break;
		case GT_8BIT:
			xpix = x / 4;
			xdot = x % 4;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x;
		break;
		case GT_15BIT:
		case GT_16BIT:
			xpix = x / 2;
			xdot = x % 2;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*2;
			break;
		case GT_24BIT:
			xpix = (x / 4) * 3;
			xdot = x % 4;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*3;
			break;
		case GT_32BIT:
			xpix = x;
			xdot = 0;
			ypix = y;
			ydot = 0;
			fb_width = dpy->mode.request.virt.x*4;
			break;
			
		default:
		INTERNALERROR;
	}
	offset = xpix + ypix * fb_width / 4;
	
	ENTER_CRITICAL;

	if (old_offset != offset) {
		outbCRT(offset >> 8, CRT_STARTADDR_H);
		outbCRT(offset, CRT_STARTADDR_L);
		outbCRT( ((offset >> 16) & CR33_DISPSTART_MASK)
		        | (inbCRT(CRT_EXT_ADDR) &~ CR33_DISPSTART_MASK),
		        CRT_EXT_ADDR);
		old_offset = offset;
	}

	if (last_ydot != ydot) {
		outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
		last_ydot = ydot;
	}

	if (last_xdot != xdot) {

		register uint hpan;

		if (dpy->mode.request.graphtype == GT_TEXT16 &&
		      dpy->mode.request.dpp.x == 9) {
			hpan = (xdot == 0) ? 0x08 : (xdot-1) & AR13_DATA_MASK;
		} else {
			hpan = (xdot & AR13_DATA_MASK) << 1;
		}

		WAIT_HVRETRACE();

		outbATC(hpan, ATC_HPAN);
		outb( ATCI_ENABLEDISPLAY, ATCI);
		last_xdot = xdot;
	}
	
	LEAVE_CRITICAL;

}

static void et6000_fb_set_split(struct kgi_display *dpy, int line)
{
	register uchar foo;

	enable_ext();
	outbCRT(line & 0xFF, CRT_LINECOMPARE);
	foo = (line & 0x0100)
		? inbCRT(0x07) | CR07_LINECOMPARE_B8
		: inbCRT(0x07) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	foo = (line & 0x0200)
		? inbCRT(0x09) | CR09_LINECOMPARE_B9
		: inbCRT(0x09) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);

	foo = (line & 0x0400)
		? inbCRT(CRT_OVRFLW_V) | CR35_LINECOMPARE_B10
		: inbCRT(CRT_OVRFLW_V) & ~CR35_LINECOMPARE_B10;
	outbCRT(foo, CRT_OVRFLW_V);
	disable_ext();
}

#define	et6000_fb_set_frame	NULL	/* !!! implement! */

#endif /* _ET6000_FB_INC */
