/* ---------------------------------------------------------------------------
**	generic fixed clock chip driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
**
**	$Log: fixed.inc,v $
**	Revision 1.11  1997/12/14 21:55:53  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.10  1997/11/08 15:12:40  becka
**	Namespace cleanup ... probably broke all and everything ...
**	
**	Revision 1.9  1997/10/03 11:26:51  seeger
**	major:	overall cleanup to GGI coding style.
**	
*/

#include <ggi/maintainers.h>
#undef	MAINTAINER
#define	MAINTAINER	Steffen_Seeger
#define	COMMON_REV	"$Revision: 1.11 $"

#if !(CLOCKVERSION == 01) 

	#warning	kgi_clock data structures changed, check this code !!
#endif

#define	FREQ	clock.clock.freq
#define	INVALIDCLK	-1

/*
**	KGI clock driver interface
*/

int kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV ",  common code:" COMMON_REV);
	
	card->clock = init_clock();

	return card->clock ? EOK : -E(CLOCK, UNKNOWN);
}

void kgim_clock_done(struct kgi_display *dpy)
{
	done_clock();
	NOTICE(DRIVER_NAME " unloaded");
}

int kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	 enum kgi_tm_cmd cmd)
{
	ggi_sint i;
	ggi_sint newfreq = (cmd == CMD_PROPOSE) 
			? 0x7FFFFFFF
			: mode->dclk * mode->clk.mul / mode->clk.div;
	ggi_sint best = 0;
	ggi_sint delta = FREQ[best] - newfreq;
	
	for (i = 1; (i < MAX_FIXED_CLOCKS) && FREQ[i]; i++)  {

		if (abs(FREQ[i] - newfreq) < abs(delta)) {

			best  = i;
			delta = FREQ[best] - newfreq;
		}
	}

	DEBUG1("best == %i, %i Hz, delta = %i\n", best, FREQ[best], delta);

	switch (cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALIDCLK;
		return EOK;

	case CMD_LOWER:

		if ((delta > 0) && (abs(delta) >= (2*newfreq/100))) {

			best++;
			if ((best == MAX_FIXED_CLOCKS) || !FREQ[best]) {

				return -E(CLOCK, UNKNOWN);
			}
		}

		mode->dclk = FREQ[best] * mode->clk.div / mode->clk.mul;
		DEBUG1("request for %i Hz, did %i Hz", newfreq, FREQ[best]);
		return EOK;

	case CMD_RAISE:

		if ((delta < 0) && (abs(delta) >= (2*newfreq/100))) {

			best--;
			if (best < 0) {

				return -E(CLOCK, UNKNOWN);
			}
		}

		mode->dclk = FREQ[best] * mode->clk.div / mode->clk.mul;
	
		DEBUG1("request for %i Hz, did %i Hz", newfreq, FREQ[best]);
		return EOK;

	case CMD_CHECK:

		if ((abs(delta) < (2*newfreq/100)) && FREQ[best]) {

			mode->clock = best;
			return EOK;

		} else {

			mode->clock = INVALIDCLK; 
			DEBUG2("invalid clock request");
			return -E(CLOCK, UNKNOWN);
		}

	default:			/* this should not happen ... */
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}

void kgim_clock_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	if (mode->clock == INVALIDCLK) {

		INTERNAL_ERROR;
	}

	CLK_out8(dpy, clkindex[mode->clock],0);
}

int kgim_clock_command(struct kgi_graphic *graph,ggi_uint cmd,ggi_uint iobuff)
{
	return -EINVAL;
}
