/* ---------------------------------------------------------------------------
**	Cirrus Logic CL-GD546x (Laguna) internal clock driver
** ---------------------------------------------------------------------------
**
**	Copyright	1997 Rodolphe ortalo
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
**	This code was derived from the following sources of information:
**
**	Laguna Family VisualMedia(TM) Accelerators CL-GD546X (Cirrus Logic)
**		Software Technical Reference Manual, 2nd edition
**		Hardware Technical Reference Manual, 2nd edition
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	"Rodolphe Ortalo, [ortalo@laas.fr]"
#define	DRIVER_NAME	"CL-GD546x (Laguna) Clock driver"
#define	DRIVER_REV	"$Revision: 1.7 $"

#undef	DEBUG
#ifdef DEBUG
	#define DEBUG_LEVEL DEBUG
#endif

#include <ggi/module.h>

#include "pll.h"

#include "Cirrus/546x.h"
/* Ugly ? -- ortalo */
extern int CLIO_in8(int);
extern void CLIO_out8(int,int);

static struct kgi_clock clock =
{
	"Cirrus Logic",			/* manufacturer		*/
	"CL-GD546x (Laguna)",		/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/

	CLK_TYPE_PROG,			/* clock type		*/
	{{
		400000, 134000000,	/* clock range		*/
	}}
};

/* PLL parameters 
** The output frequency is calculated as
** fout = (fref * factor.mul/factor.div * dividend/divisor) >> post-VCO;
*/

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 1, 1 },			/* factor		 	*/
	{ 1000000, 120000000 },		/* fvco range			*/
	{ 0, 1 },			/* post-VCO divider range	*/
	/* We could use pixel doubling here, but apparently, only for
	** graphic modes 8BPP and 16BPP, pp.S5-37,H5-8 (?) -- ortalo
	*/
	#warning could be 0 low here ? -- ortalo
	{ 1, 127 },			/* dividend range		*/
	{ 1, 127 }			/* divisor range		*/
};


#define	INVALID_COMMAND	((ggi_uint) -1)

/*	A clock command word is as follows:
**
**	[0:6]	s->mul	
**	[8:14]	s->div
**	[16]	s->p
*/
static ggi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->mul < 128);
	ASSERT(s->div < 128);
	ASSERT(s->p < 2);
	DEBUG2("Building clock command for f=%d,fvco=%d with mul=%d,div=%d,p=%d\n",
	       s->f,s->fvco,s->mul,s->div,s->p);
	/* A ggi_uint value that encodes the state */
	return	((s->mul) | (s->div << 8) | (s->p << 16));
}

static void pll_set(struct kgi_display *dpy, ggi_uint clock_cmd)
{
	/* select VCLK3 as clock (11)*/
	CLIO_out8(MISC_OUTPUT,CLIO_in8(MISC_INPUT) | MISC_VCLK3);
	/* sets the numerator */
	CLIO_out8(SR1E_VCLK3_N, clock_cmd & VCLK_N_MASK);
	/* sets the denominator and post-scalar */
	CLIO_out8(SR0E_VCLK3_D_P, ((clock_cmd >> 7) & VCLK_D_MASK) |
			((clock_cmd >> 16) & VCLK_P_MASK) );
	return;
}

/*
**	KGI module interface and PLL things
*/

#include "pll.inc"





