/* -----------------------------------------------------------------------------
**	TI TVP 3026 ramdac driver
** -----------------------------------------------------------------------------
**
**	Copyright (C)	1997		Jonas Borgstrom
**			1997		Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/
/* If you have a ramdac with a higher Mhz then 135 MHz, then use
 * the modeule parameter 'dacmhz', ex: 'insmod ggidrv.o dacmhz=220'
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define	DRIVER_NAME	"TVP 3026 DAC driver"
#define	DRIVER_REV	"$Revison: 1.10 $"

#define	DEBUG_LEVEL	0	/* do silent checking */

#undef	DEBUG

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#define DAC_PRECISION	8	/* do 8 bit R,G,B resolution */

#include <ggi/module.h>
#include "TI/tvp3026.h"

#define	SUPPORTED (GM_TEXT16 | GM_4BIT | GM_8BIT \
		  | GM_15BIT | GM_16BIT | GM_32BIT)


#define TVP3026_135 0
#define TVP3026_175 1
#define TVP3026_220 2
#define TVP3026_250 3
#define DAC_UNKNOWN -1

/* 
 * Module parameter
 */
extern ggi_uint dacmhz;

static struct kgi_ramdac tvp3026_info[4] =
{
       {
		"Texas Instruments",	/* manufacturer		*/
		"TVP 3026-135",		/* model		*/
		SUPPORTED,			/* supported modes	*/
		FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
		{ 0x0FFF, 0x0FFF },		/* maximum resolution	*/

		DAC_8BIT_PB | DAC_16BIT_PB |	/* device type		*/
		DAC_32BIT_PB | DAC_64BIT_PB | DAC_VRAM,

		 {        0, 110000000 },	/* normal mode		*/
		 {        0,         0 },	/* 2xclk mode		*/

		{{	  0,  75000000 },	/* 	GT_TEXT		*/
		 {        0,         0 },	/*	GT_TEXT32	*/
		 {        0, 135000000 },	/* 	GT_1BIT		*/
		 {	  0, 135000000 },	/* 	GT_4BIT		*/
		 {        0, 135000000 },	/* 	GT_8BIT		*/	
		 {	  0, 135000000 },	/* 	GT_15BIT	*/
		 { 	  0, 135000000 },	/* 	GT_16BIT	*/
		 {    	  0, 135000000 },	/* 	GT_24BIT	*/
	 	 {	  0, 110000000 }}	/* 	GT_32BIT	*/
	},

	{
		"Texas Instruments",	/* manufacturer		*/
		"TVP 3026-175",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

		 DAC_8BIT_PB | DAC_16BIT_PB |	/* device type		*/
		 DAC_32BIT_PB | DAC_64BIT_PB | DAC_VRAM,

		  {       0, 110000000 },	/* limit in normal mode	*/
		  {       0,         0 },	/* limit in 2xclk mode	*/
 
		 {{	  0,  75000000 },	/* 	GT_TEXT		*/
		  {       0,         0 },	/*	GT_TEXT32	*/
		  {       0, 175000000 },	/* 	GT_1BIT		*/
		  {	  0, 175000000 },	/* 	GT_4BIT		*/
		  {       0, 175000000 },	/* 	GT_8BIT		*/	
		  {	  0, 175000000 },	/* 	GT_15BIT	*/
		  { 	  0, 175000000 },	/* 	GT_16BIT	*/
		  {       0, 175000000 },	/* 	GT_24BIT	*/
		  {	  0, 110000000 }}	/* 	GT_32BIT	*/
	},

	{
		"Texas Instruments",	/* manufacturer		*/
		"TVP 3026-220",		/* model		*/
		SUPPORTED,			/* supported modes	*/
		FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
		FL_DAC_2XCLOCK,
		{ 0x0FFF, 0x0FFF },		/* maximum resolution	*/

		 DAC_8BIT_PB | DAC_16BIT_PB |	/* device type		*/
		 DAC_32BIT_PB | DAC_64BIT_PB | DAC_VRAM,

	     {       0, 110000000 },		/* normal mode		*/
	     {       0,         0 },
		{{	  0,  75000000 },	/* 	GT_TEXT		*/
		 {        0,         0 },	/*	GT_TEXT32	*/
		 {        0, 220000000 },	/* 	GT_1BIT		*/
		 {	  0, 220000000 },	/* 	GT_4BIT		*/
		 {        0, 220000000 },	/* 	GT_8BIT		*/
		 {	  0, 220000000 },	/* 	GT_15BIT	*/
		 { 	  0, 220000000 },	/* 	GT_16BIT	*/
		 {        0, 220000000 },	/* 	GT_24BIT	*/
		 {	  0, 110000000 }}	/* 	GT_32BIT	*/
	},

	{
		"Texas Instruments",	/* manufacturer		*/
		 "TVP 3026-250",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		 FL_DAC_2XCLOCK,
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

		 DAC_8BIT_PB | DAC_16BIT_PB |	/* device type		*/
		 DAC_32BIT_PB | DAC_64BIT_PB | DAC_VRAM,

		  {       0, 110000000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  85000000 },	/* 	GT_TEXT		*/
		  {       0,         0 },	/*	GT_TEXT32	*/
		  {       0, 250000000 },	/* 	GT_1BIT		*/
		  {	  0, 250000000 },	/* 	GT_4BIT		*/
		  {       0, 250000000 },	/* 	GT_8BIT		*/	
		  {	  0, 259000000 },	/* 	GT_15BIT	*/
		  { 	  0, 250000000 },	/* 	GT_16BIT	*/
		  {       0,         0 },	/* 	GT_24BIT	*/
		  {	  0, 111000000 }}	/* 	GT_32BIT	*/
	}
};

/*
**	driver state info
*/

struct tvp3026_private 
{
	ggi_uint	cur_hotx, cur_hoty;	/* cursor hot spot	*/
	ggi_uint	cur_x, cur_y;		/* current position	*/
	ggi_uint	rev;			/* revision		*/
};

#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y

#define	LCLK(m,d)	((mode->lclk.mul == m) && (mode->lclk.div == d))

#define	GET_PRIVATE	struct tvp3026_private *private = KGIM_PRIVATE(dpy,ramdac)

/*
**	helper functions
*/

#include "IBM/clut.inc"		/* use dac_set_clut() from normal VGA */

static inline void EXT_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	DAC_out8(dpy, reg, DAC_EXT_ADDR);
	DAC_out8(dpy, val, DAC_EXT_DATA);
}

static inline unsigned char EXT_in8(struct kgi_display *dpy, uint8 reg)
{
	DAC_out8(dpy, reg, DAC_EXT_ADDR);
	return DAC_in8(dpy, DAC_EXT_DATA);
}

static void dac_set_pointer_shape(struct kgi_display *dpy, 
	struct kgi_pointer64x64 *ptr)
{
	GET_PRIVATE;
	ggi_uint i;
	uint8 ext06;

	/*	update private state
	*/
        ext06 = EXT_in8(dpy, 0x06);
	ext06 &= ~DAC_EXT06_CursorMask;
	switch (ptr->mode) {

	case pmWindows:
		ext06 |= DAC_EXT06_CursorXGA;
		break;

	case pmX11:
		ext06 |= DAC_EXT06_CursorX;
		break;

	case pmThreeColor:
		ext06 |= DAC_EXT06_CursorThreeColor;
		break;

	default:
		INTERNAL_ERROR;
	}
	CUR_HOTX = ptr->hotx;
	CUR_HOTY = ptr->hoty;
        ext06 &= 0xF3; /* Reset A9, A8 */

	EXT_out8(dpy, ext06, 0x06);
        DAC_out8(dpy, 0x00, 0x00); /* Reset A0 - A7 */
	DEBUG1("set_pointer_shape: hot spot (%i,%i), EXT06 %i", \
		CUR_HOTX, CUR_HOTY, ext06);

	/*	set pattern
	*/
        for (i = 0;i < 512; i++) {

        	DAC_out8(dpy, ptr->xor_mask[i], DAC_CURSOR_RAM_DATA);
        }
        for (i = 0;i < 512; i++) {

        	DAC_out8(dpy, ptr->and_mask[i], DAC_CURSOR_RAM_DATA);
        }

	/*
	**	set colors
	*/
	DAC_out8(dpy, 1, DAC_CW_INDEX);
	for (i = 0; i < 3; i++) {

		DAC_out8(dpy, ptr->col[i].r >> SHIFT, DAC_C_DATA);
		DAC_out8(dpy, ptr->col[i].g >> SHIFT, DAC_C_DATA);
		DAC_out8(dpy, ptr->col[i].b >> SHIFT, DAC_C_DATA);
	}
}

static void dac_show_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_PRIVATE;

	if ((CUR_X != x) || (CUR_Y != y)) {

		CUR_X = x;
		x += 64 - CUR_HOTX + dpy->mode.x.total - dpy->mode.x.blankend;

		CUR_Y = y;
		y += 64 - CUR_HOTY + dpy->mode.y.total - dpy->mode.y.blankend;

		DAC_out8(dpy, x,      DAC_CURSOR_XL);
		DAC_out8(dpy, x >> 8, DAC_CURSOR_XH);
		DAC_out8(dpy, y,      DAC_CURSOR_YL);
		DAC_out8(dpy, y >> 8, DAC_CURSOR_YH);
	}
}

static void dac_hide_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;

	CUR_X = 0x0FFF;
	CUR_Y = 0x0FFF;

	DAC_out8(dpy, 0xFF, DAC_CURSOR_XL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_XH);
	DAC_out8(dpy, 0xFF, DAC_CURSOR_YL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_YH);
}

/*
**	KGI interface
*/

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	uint8 status = -1;
	ggi_uint dac_type;
	struct tvp3026_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

 	if (EXT_in8(dpy, 0x3F) != DAC_EXT3F_TVP3026_ID) {

		DEBUG0("unknown chip ID (%.2x).", status);
		return -E(RAMDAC, UNKNOWN);
	}

	if (!(private = kgi_malloc_small(sizeof(*private), GFP_KERNEL))) {

		ERROR("failed to allocate private data!");
		return -ENOMEM;
	}
   
	KGIM_PRIVATE(dpy,ramdac) = private;
	memset(private, 0, sizeof(*private));

	private->rev = EXT_in8(dpy, DAC_EXT_Revision);
	DEBUG0("TVP 3026 rev %.2x.%.2x detected.",
		(private->rev & 0xf0) >> 4, (private->rev & 0x0f));
   
	switch(dacmhz) {
		case 175:
			dac_type = TVP3026_175;
			break;
	 	case 220:
	   		dac_type = TVP3026_220;
	   		break;
	 	case 250:
	   		dac_type = TVP3026_250;
	   		break;
	 	default:
	   		dac_type = TVP3026_135;
	   		dacmhz = 135;
	};
	card->ramdac = &tvp3026_info[dac_type];
	DEBUG0("Using %d MHz mode", dacmhz);
   
	DEBUG1("  .. %s %s loaded.", card->ramdac->manufact, \
		card->ramdac->model);
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	kgi_free_small(KGIM_PRIVATE(dpy,ramdac));
	KGIM_PRIVATE(dpy,ramdac) = NULL;
	NOTICE(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode, 
			enum kgi_tm_cmd cmd)
{
	int div;
	int newclock = mode->dclk;
	struct kgi_ramdac *dac = dpy->card->ramdac;

	switch (cmd) {

	case CMD_PROPOSE:

		switch(mode->request.graphtype) {

		case GT_TEXT16:
			div =  1;
			break;

		case GT_4BIT:
			div = 16;
			break;

		case GT_8BIT:
			div =  8;
			break;

		case GT_15BIT:	/* fall through	*/
		case GT_16BIT:
			div =  4;
			break;

		case GT_32BIT:
			div =  2;
			break;

		case GT_24BIT:
			div =  8;
			break;

		default:
			DEBUG2("could not handle mode %x",
				mode->request.graphtype);
			return -E(RAMDAC, UNKNOWN);
		}

		mode->lclk.mul = mode->clk.mul = mode->clk.div = 1;
	      	if (mode->request.graphtype == GT_24BIT) {

			mode->lclk.mul = 3;
		}
		mode->lclk.div = div;
		DEBUG2("%i:%i multiplex mode", mode->lclk.div, mode->lclk.mul);

		mode->flags &= ~FL_DAC_2XCLOCK;
		mode->flags |= FL_DAC_8BIT;
		if ((mode->request.graphtype == GT_15BIT) || 
			(mode->request.graphtype == GT_16BIT) ||
			(mode->request.graphtype == GT_24BIT) ||
			(mode->request.graphtype == GT_32BIT)) {

			mode->flags |= FL_DAC_GAMMA;
		}
 		return EOK;

	case CMD_LOWER:

		if ((newclock < dac->clock[mode->request.graphtype].min) ||
			(newclock < dac->normal.min)) {

			printk("%i Hz DCLK is too low\n", newclock);
			return -E(RAMDAC, UNKNOWN);
		}

		if (dac->clock[mode->request.graphtype].max < newclock) {

    			newclock = dac->clock[mode->request.graphtype].max;
		}

		DEBUG2("dclk = %i", newclock);
		mode->dclk   = newclock;
		return EOK;

	case CMD_RAISE:

		if ((newclock > dac->clock[mode->request.graphtype].max) ||
			(newclock > dac->normal.max)) {

			printk("%i Hz DCLK is too high\n", newclock);
			return -E(RAMDAC, UNKNOWN);
		}

		if (newclock < dac->clock[mode->request.graphtype].min) {

			newclock = dac->clock[mode->request.graphtype].min;
		}

		if (newclock < dac->normal.min) {

			newclock = dac->normal.min;
		}

		mode->dclk = newclock;
		return EOK;

     case CMD_CHECK:

		switch(mode->request.graphtype) {

		case GT_TEXT16:
			if (! LCLK(1,1)) {

				return -E(RAMDAC, NOSUP);
			}
			break;

		case GT_4BIT:
			if (! (LCLK(1,16) || LCLK(1,8) || LCLK(1,4) ||
				LCLK(1,2))) {

				return -E(RAMDAC, NOSUP);
			}
			break;

		case GT_8BIT:
			if (! (LCLK(1,8) || LCLK(1,4) || LCLK(1,2) ||
				LCLK(1,1))) {
				return -E(RAMDAC, NOSUP);
			}
			break;
 
		case GT_15BIT:
		case GT_16BIT:
			if (! (LCLK(1,4) || LCLK(1,2) || LCLK(1,1))) {
				return -E(RAMDAC, NOSUP);
			}
			break;
  
		case GT_32BIT:
			if (! (LCLK(1,2) || LCLK(1,1))) {
				return -E(RAMDAC, NOSUP);
			}
			break;
  
		case GT_24BIT:
			if (! (LCLK(3,8) || LCLK(3,4))) {

				return -E(RAMDAC, NOSUP);
			}
			break;

		default:
			DEBUG2("could not handle mode %x",
				mode->request.graphtype);
			return -E(RAMDAC, UNKNOWN);
		}

                if (((mode->request.graphtype == GT_TEXT16) &&
			(mode->dclk < dac->normal.min)) ||
			mode->dclk < dac->clock[mode->request.graphtype].min) {

			DEBUG2("dclk < norm.min\n");
			return -E(RAMDAC, CLK_LOW_LIMIT);
                }
	   
                if (((mode->request.graphtype == GT_TEXT16) &&
			(mode->dclk > dac->normal.max)) ||
			mode->dclk > dac->clock[mode->request.graphtype].max) {

			DEBUG2("dclk > norm.max\n");
			return -E(RAMDAC, CLK_HIGH_LIMIT);
                }
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	GET_PRIVATE;
        unsigned char EXT0F, EXT18, EXT1A, EXT1E, EXT19, EXT1D, EXT2B;
	enum kgi_sync synctype = dpy->card->monitor->synctype;
	uint lm, ln, lp, z, lq;

	switch(mode->request.graphtype) {

	case GT_TEXT16:
		EXT18 = 0x80;
		EXT19 = 0x98;
                EXT0F = 0x06;
                EXT1A = 0x07;
                EXT1E = DAC_EXT1E_6n8Disable | DAC_EXT1E_8BitEnable;
                EXT2B = 0x14;
		break;

	case GT_4BIT:
		EXT18 = 0x80;
		EXT0F = 0x06;
                EXT1A = 0x00;
                EXT1E = DAC_EXT1E_6n8Disable | DAC_EXT1E_8BitEnable;
                EXT2B = 0x14;

		switch (mode->lclk.div) {

		case  2:
			EXT19 = 0x41;
			break;

		case  4:
			EXT19 = 0x42;
			break;

		case  8:
			EXT19 = 0x43;
			break;

		case 16:
			EXT19 = 0x44;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
		break;

	case GT_8BIT:
		EXT18 = 0x80;
                EXT0F = 0x06;
                EXT1A = 0x25;
                EXT1E = DAC_EXT1E_6n8Disable | DAC_EXT1E_8BitEnable;
                EXT2B = 0x10;

		switch (mode->lclk.div) {

		case  1:
			EXT19 = 0x49;
			break;

		case  2:
			EXT19 = 0x4A;
			break;

		case  4:
			EXT19 = 0x4B;
			break;

		case  8:
			EXT19 = 0x4C;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
		break;
 
	case GT_16BIT:
	case GT_15BIT:
                EXT0F = 0x07;
                EXT1A = 0x15;
                EXT1E = 0x2C;
                EXT2B = 0x1E;

		EXT18 = (mode->request.graphtype == GT_15BIT) ? 0x04 : 0x05;

		switch (mode->lclk.div) {

		case  1:
			EXT19 = 0x52;
			break;

		case  2:
			EXT19 = 0x53;
			break;

		case  4:
			EXT19 = 0x54;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
		break;
  
	case GT_24BIT:
		EXT18 = 0x16;
                EXT0F = 0x07;
                EXT1A = 0x25;
                EXT1E = 0x2C;
                EXT2B = 0x1E;

		switch (mode->lclk.div) {

		case  4:
			EXT19 = 0x5B;
			if (private->rev > 0x20) {

				EXT0F = 0x08;
			}
			break;

		case  8:
			EXT19 = 0x5C;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
		break;

	case GT_32BIT:
		EXT18 = 0x06;
                EXT0F = 0x07;
                EXT1A = 0x05;
                EXT1E = 0x2C;
                EXT2B = 0x1E;

		switch (mode->lclk.div) {

		case 1:
			EXT19 = 0x5B;
/*	???		if (private->rev > 0x20) {
**	???
**	???			EXT0F = 0x08;
**	???		}
*/
                        break;
		case 2:
			EXT19 = 0x5C;
                        break;
		default:
			INTERNAL_ERROR;
			return;
		}
		break;

	default:
		INTERNAL_ERROR;
		DEBUG2("could not handle mode %x", mode->request.graphtype);
		return;
	}	
	EXT_out8(dpy, 0, 0x06);
        EXT_out8(dpy, EXT0F, 0x0F);
	EXT_out8(dpy, EXT18, 0x18);
	EXT_out8(dpy, EXT19, 0x19);
        EXT_out8(dpy, EXT1A, 0x1A);
        EXT_out8(dpy, EXT1E, 0x1E);
        EXT_out8(dpy, EXT2B, 0x2B);

/*	#warning hardware cursor in interlaced modes!
	EXT_out8(dpy, DAC_EXT06_DirectCCR, 0x06);*/
	DAC_out8(dpy, DAC09_CursorOff, 0x09);

	/*	set general control register
	*/
	EXT1D = 0;

	if (synctype & (SYNC_ON_RED | SYNC_ON_BLUE)) {

		INTERNAL_ERROR;
		return;
	}
	if (synctype & SYNC_ON_GREEN) {

		EXT1D |= DAC_EXT1D_SyncEnable;
	}
	if (synctype & SYNC_PEDESTRAL) {

		EXT1D |= DAC_EXT1D_BlankPedestral;
	}
	EXT_out8(dpy, EXT1D, 0x1D);

/*        if(mode->magnify.x == 2) {

                printk("maginfy = %d\n", mode->magnify.x);
        	EXT_out8(dpy, (EXT_in8(dpy, DAC_EXT_ColKeyControl) & 0x1f)
        		| (1 << DAC_EXT38_ZoomShift),
        		DAC_EXT_ColKeyControl);
        } else {

        	EXT_out8(dpy, EXT_in8(dpy, DAC_EXT_ColKeyControl) & 0x1f,
	        	DAC_EXT_ColKeyControl);
        }*/

	EXT_out8(dpy, 0, DAC_EXT_PalettePage);

	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);

	CUR_X = -1;
	CUR_Y = -1;
	DAC_out8(dpy, 0xFF, DAC_CURSOR_XL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_XH);
	DAC_out8(dpy, 0xFF, DAC_CURSOR_YL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_YH);

	if (mode->request.graphtype == GT_24BIT) {

		#warning handle revision A/B properly!
		/* ln:lm = ln:3 */
		lm = 65 - 3;
		ln = 65 - mode->lclk.div;

		/* Note: this is actually 100 * z for more precision */
		z = (11000000 * (65 - ln)) / ((mode->dclk / 1000) * (65 - lm));
		
	} else {

		ASSERT(mode->lclk.mul == 1);
		ln = 65 - 4 * mode->lclk.div;
		lm = 61;

		/*	now calculate z. This is really z * 100
		*/
		if ((mode->clk.mul == 1) && (mode->clk.div == 1)) {

			z = 11000000 / (mode->dclk / 1000) ;

		} else {

			z = 11000000 / (mode->dclk / mode->lclk.div / 1000);
		}
	}

	if (z > 1600) {

		/*	smallest q greater (z - 16) / 16
		*/
		lp = 3;
		lq = ((z-1600) / 1600) + 1;

	} else {

		/*	largest p less then log2(z)
		*/
		for (lp = 0; z > (200 << lp); lp++) {
		}
		lq = 0;
	}

	DEBUG1("Loop PLL { m, n, q } = { %i, %i, %i }", lm, ln, lq);

	if ( mode->request.graphtype == GT_24BIT ) {

		ln = (ln & 0x3f) | 0x80;
		lm = (lm & 0x3f) | 0x80;
		lp = (lp & 0x03) | 0xf8;

	} else {

		ln = (ln & 0x3f) | 0xc0;/*0x80;*/
		lm = (lm & 0x3f);
		lp = (lp & 0x03) | 0xf0;
	}

	DAC_out8(dpy, DAC_EXT_PLLAddress, DAC_EXT_ADDR);
	DAC_out8(dpy, 0x00, DAC_EXT_DATA);
      
	DAC_out8(dpy, DAC_EXT_LClkData, DAC_EXT_ADDR);
	DAC_out8(dpy, ln, DAC_EXT_DATA);
	DAC_out8(dpy, lm, DAC_EXT_DATA);
	DAC_out8(dpy, lp, DAC_EXT_DATA);

	if (mode->request.graphtype == GT_TEXT16) {
      
		EXT_out8(dpy, 0x18, 0x39);

	} else {

		EXT_out8(dpy, lq | 0x38, 0x39);
	}

        dpy->set_clut = dac_set_clut;
	dpy->set_pointer_shape  = dac_set_pointer_shape;
        dpy->show_pointer = dac_show_pointer;
        dpy->hide_pointer = dac_hide_pointer;
	dpy->undo_pointer = NULL;
}

int kgim_ramdac_command (struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/

	ggi_clut clut;

	switch(cmd) {

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data=(ggi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	case RAMDAC_GETCLUT:
		clut.size=256;
		clut.data=(ggi_color *)(graph->io_buf);
		graph->io_res=graph->io_buf;
		dac_get_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	}

	/*	This driver can't do this. Maybe a simpler command will do.
	*/
	return -E(RAMDAC,NOSUP_ALWAYS_LOWER);	
}
