/*\
|*| ----------------------------------------------------------------------------
|*|	GGI display and drawable definitions
|*| ----------------------------------------------------------------------------
|*|	NOTE: NEVER INCLUDE THIS FILE DIRECTLY! ONLY USE <ggi.h>!
|*| ----------------------------------------------------------------------------
|*|	Copyright (C) 1996 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
\*/

typedef struct pixmap
{
	struct display	*dpy;		/* display this is displayed	*/
#ifndef __KERNEL__
	struct pixmap	*parent;	/* parent if it is a childmap	*/
	struct pixmap	*next;		/* circular list of brothers	*/
	struct pixmap	*childs;	/* entry to child pixmap list	*/

	ggi_coord	offset;		/* offset from parent origin	*/
#endif
	ggi_size	size;		/* size of the (child)map	*/
	ggi_maptype	bpp;		/* buffer layout		*/
	ggi_pixel	*pbuf;		/* pixel value buffer		*/
	ggi_zvalue	*zbuf;		/* z value buffer		*/
	struct fb_ops	*ops;
} pixmap;

typedef struct fb_ops
{
	void	(*move)(pixmap *fb, ggi_uint dst, ggi_uint src, ggi_uint cnt);
	void	(*set) (pixmap *fb, ggi_uint s, pixel val, ggi_uint cnt);
	void	(*invert) (pixmap *fb, ggi_uint start, ggi_uint end);
	void	(*update) (FBARGD, ggi_uint pos, pixel val);
	void	(*flush)  (pixmap *fb);
	void	(*copy_from)(pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt);
	void	(*copy_to)(pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt);

	pixel	(*a_to_p) (pixmap *fb, attr a);
	attr	(*p_to_a) (pixmap *fb, pixel val);
	pixel	(*f_to_p) (pixmap *fb, ggi_uint pos);
	ggi_uint	(*p_to_f) (pixmap *fb, pixel val);

#ifndef __KERNEL__

	void	(*points) 	(FBARGD, ggi_point_data *data, ggi_uint cnt);
	void	(*lines) 	(FBARGD, ggi_line_data  *data, ggi_uint cnt);
	void	(*rectangles)	(FBARGD, ggi_rect_data  *data, ggi_uint cnt);
	void	(*fillrects)	(FBARGD, ggi_rect_data  *data, ggi_uint cnt);
	void	(*arcs)		(FBARGD, ggi_arc_data   *data, ggi_uint cnt);
	void	(*fillarcs)	(FBARGD, ggi_arc_data   *data, ggi_uint cnt);

	void	(*polygon)  (FBARGD, ggi_point_data *data, ggi_uint cnt, int mode);
	void	(*fillpoly) (FBARGD, ggi_point_data *data, ggi_uint cnt, int mode);
#endif
} fb_ops;

#ifdef __KERNEL__
extern inline void fb_move(pixmap *fb, ggi_uint dst, ggi_uint src, ggi_uint cnt)
{ (fb->ops->move)(fb, dst, src, cnt); }

extern inline void fb_memset(pixmap *fb, ggi_uint s, ggi_uint val, ggi_uint cnt)
{ (fb->ops->set)(fb, s, val, cnt); }

extern inline void fb_invert(pixmap *fb, ggi_uint start, ggi_uint end)
{ (fb->ops->invert)(fb, start, end); }

extern inline void fb_update(FBARGD, ggi_uint pos, ggi_uint val)
{ (fb->ops->update)(FBARGC, pos, val); }

extern inline void fb_flush(pixmap *fb)
{ (fb->ops->flush)(fb); }

extern inline void fb_copy_from(pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{ (fb->ops->copy_from)(fb, buf, start, cnt); }

extern inline void fb_copy_to(pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{ (fb->ops->copy_to)(fb, buf, start, cnt); }

extern inline pixel fb_attr_to_pixel(pixmap *fb, attr a)
{ return (fb->ops->a_to_p)(fb, a); }

extern inline attr fb_pixel_to_attr(pixmap *fb, pixel val)
{ return (fb->ops->p_to_a)(fb, val); }

extern inline pixel fb_fpos_to_pixel(pixmap *fb, ggi_uint c)
{ return (fb->ops->f_to_p)(fb, c); }

extern inline ggi_uint fb_pixel_to_fpos(pixmap *fb, pixel val)
{ return (fb->ops->p_to_f)(fb, val); }
#endif
