/* 
   Banked Access Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1997 Brian S. Julin -- bri@calyx.com

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <asm/page.h>

#define BANKFB __localafb
#define BANKFB_LEN __localafb_len
#include "ggi-dl.h"
#include <ggi/ggi_commands.h>

/* 
 * We are guaranteed to have been checked already...
 */

#define BANK_PAGE_SIZE (32 * 1024) /* Later get from new ioctl */
#define MMAP_PAGE_SIZE ((PAGE_SIZE > BANK_PAGE_SIZE) ? PAGE_SIZE : BANK_PAGE_SIZE)

unsigned char *__localrfb=NULL;  /* Read only bank */
unsigned char *__localwfb=NULL;  /* Write-only bank */
unsigned char *__localafb=NULL;  /* Bank that anticipate ascending access */
unsigned char *__localdfb=NULL;  /* Bank that anticipates descending access */
static int __localafb_len;
static int __localdfb_len;
static int __localrfb_len;
static int __localwfb_len;
static ggi_gc *_ggiGC=NULL;

int GGIdlinit(ggi_visual *visual,const char *version)
{
  
  if (VIS_MODE(visual)) {

    __localafb_len =
    __localdfb_len =
    __localrfb_len =
    __localwfb_len =
      (((VIS_MODE(visual)->virt.x * VIS_MODE(visual)->virt.y) 
		    & ~(MMAP_PAGE_SIZE - 1)) + MMAP_PAGE_SIZE);

    DPRINT("afb is this big: %d\n",__localafb_len);
    DPRINT("dfb is this big: %d\n",__localdfb_len);
    DPRINT("rfb is this big: %d\n",__localrfb_len);
    DPRINT("wfb is this big: %d\n",__localwfb_len);
  }
  else {
    printf("(fatal) NULL mode provided...\n");
    return(-1);
  };


  __localafb=(unsigned char *)mmap(NULL,
				   __localafb_len,
				   PROT_READ | PROT_WRITE,
				   MAP_SHARED,
				   VIS_FD(visual),
				   MMAP_TYPE_MMIO |
				   MMAP_PER_REGION_TYPE| 
				   MMAP_PRIVATE | 4);
  if ((long)(__localafb) == -1) { __localafb = (unsigned char *)NULL; }; 


  __localdfb=(unsigned char *)mmap(NULL,
				   __localdfb_len,
				   PROT_READ | PROT_WRITE,
				   MAP_SHARED,
				   VIS_FD(visual),
				   MMAP_TYPE_MMIO |
				   MMAP_PER_REGION_TYPE| 
				   MMAP_PRIVATE | 8);
  if ((long)(__localdfb) == -1) { __localdfb = (unsigned char *)NULL; }; 
  
  
  __localrfb=(unsigned char *)mmap(NULL,
				   __localrfb_len,
				   PROT_READ,
				   MAP_SHARED,
				   VIS_FD(visual),
				   MMAP_TYPE_MMIO |
				   MMAP_PER_REGION_TYPE| 
				   MMAP_PRIVATE | 16);
  if ((long)(__localrfb) == -1) { __localrfb = (unsigned char *)NULL; }; 
  
  
  __localwfb=(unsigned char *)mmap(NULL,
				   __localwfb_len,
				   PROT_WRITE,
				   MAP_SHARED,
				   VIS_FD(visual),
				   MMAP_TYPE_MMIO |
				   MMAP_PER_REGION_TYPE| 
				   MMAP_PRIVATE | 32);
  if ((long)(__localwfb) == -1) { __localwfb = (unsigned char *)NULL; }; 
  

  /* Now attempt make-do with whatever the graphics hardware provided. */
  if (!__localrfb) {
    /* Must provide Bank1 */
    if (!__localafb) { printf("No banks\n"); return(-1); };
    /* Use Ascending Bank in leiu of real RO bank, if it exists that is. */
    __localrfb = __localafb;
  }

  if (!__localwfb) {
    /* Must provide Bank1 */
    if (!__localafb) { printf("No banks\n"); return(-1); };
    /* Use Ascending Bank in leiu of real WO bank, if it exists that is. */
    __localwfb = __localafb;
  }

  if (!__localdfb) { 
    /* Must provide Ascending Bank */
    if (!__localafb) { printf("No banks\n"); return(-1); };
    __localdfb = __localafb;  /* This won't be used often if at all */
  } 
  
  
  DPRINT("afb=%p, dfb=%p, rfb=%p, wfb=%p\n",
	 __localafb, __localdfb, __localrfb, __localwfb);

	if (VIS_GC(visual)==NULL) {
		_ggiGC=malloc(sizeof(ggi_gc));
		if (_ggiGC==NULL) {
		  if (__localafb!=NULL)
		    free(__localafb);
		  if (__localdfb!=NULL && __localdfb!=__localafb)
		    free(__localdfb);
		  if (__localrfb!=NULL && __localrfb!=__localafb)
		    free(__localrfb);
		  if (__localwfb!=NULL && __localwfb!=__localafb
		      && __localwfb!=__localdfb)
		    free(__localwfb);
		  return -1;
		}

		_ggiGC->version=0;
/*		_ggiLockInit(_ggiGC);*/
	}

	if (VIS_GC(visual)==NULL)
		VIS_GC(visual)=_ggiGC;

	/* Linear framebuffer setup */
	if (VIS_FB_LINEAR(visual) != NULL) {
	  munmap(VIS_FB_LINEAR(visual), VIS_FB_LINEAR_SIZE(visual));
	}
	VIS_FB_LINEAR(visual) = BANKFB;
	VIS_FB_LINEAR_SIZE(visual) = BANKFB_LEN;
	VIS_FB_BPP(visual) = 8;
	VIS_FB_BPC(visual) = 8;
	VIS_FB_BPA(visual) = 0;

	GGIdbinit(visual);

	/* Color mapping 
	 */

	/* Positioning is not supported...
	 */

	/* Sprites aren't supported...
	 */

	/* Generic drawing
	 */
	/*	visual->op2d->fillscreen=GGIfillscreen; */
        /*
	visual->op2d->putc=GGIputc;
	visual->op2d->puts=GGIputs; */

	visual->op2d->drawpixel=GGIdrawpixel;
	visual->op2d->_drawpixel=_GGIdrawpixel;
	visual->op2d->putpixel=GGIputpixel;
	visual->op2d->getpixel=GGIgetpixel;
	
	visual->op2d->drawhline=GGIdrawhline;
	visual->op2d->_drawhline=_GGIdrawhline;
	visual->op2d->puthline=GGIputhline;
	visual->op2d->gethline=GGIgethline;

	visual->op2d->drawvline=GGIdrawvline;
	visual->op2d->_drawvline=_GGIdrawvline;
	visual->op2d->putvline=GGIputvline;
	visual->op2d->getvline=GGIgetvline;
	/*
	visual->op2d->drawbox=GGIdrawbox; 
	visual->op2d->putbox=GGIputbox;
	visual->op2d->getbox=GGIgetbox;
	*/
	/*
	visual->op2d->drawline=GGIdrawline;
	*/

	return GGI_DL_OPCOLOR|GGI_DL_OP2D;
}

int GGIdlcleanup(ggi_visual *visual)
{
	if (__localafb!=NULL) {
	  munmap((char *)__localafb,__localafb_len);
	};
	if (__localdfb!=NULL && __localdfb!=__localafb) {
	  munmap((char *)__localdfb,__localdfb_len);
	};
	if (__localrfb!=NULL && __localrfb!=__localafb) {
	  munmap((char *)__localrfb,__localrfb_len);
	};
	if (__localwfb!=NULL && __localwfb!=__localafb
	    && __localwfb!=__localdfb) {
	  munmap((char *)__localwfb,__localwfb_len); 
	};
	
	if (VIS_GC(visual)==_ggiGC) {
	  /*		_ggiLockDestroy(VIS_GC(visual));*/
	  free(VIS_GC(visual));
	  VIS_GC(visual)=NULL;
	  _ggiGC=NULL;
	}
	
	return EOK;
}
		
#include "ggidlinit.h"
