#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <ggi/libggi.h>
#include <linux/kd.h>
#include <signal.h>
#include "evstack.h"
#include "parsearg.h"

#undef MEGA_DEBUG
#undef USE_PIPETTY

ggi_visual_t MASTER_VIS=NULL;

/* Number of termstacks to create */
#define TERMSTACKNUM 10

/* This demo takes the job of some device driver which sends data to an
   event queue. */

void dump_stack(evstack *stack,char *class)
{
	int i;
	char buffer[256];
	
	printk("$ cat /proc/%s/%08x\n",class,stack->id);
	for (i=0;;i++) {
		if (stack->ops->read(stack,i,sizeof(buffer),buffer)<0)
		  break;
		printk("%s\n",buffer);
	}
}

evstack *create_and_attach_stack(evstack *termstack,const char *name)
{
	int id;
	evstack *stk;
	char buffer[128];

	id=ev_lookup_class(name);

	if (id<0) {
		printk(KERN_ERR "Can't find stack class \"%s\"\n",
			name);
		exit(1);
	}

	/* Tell which stack to send info to */
	stk=ev_create_stack(id);
	sprintf(buffer,"attach %s",par_stack_to_string(stk));

	termstack->ops->write(termstack,buffer);

	return stk;
}

evstack *clone_and_attach_stack(evstack *termstack,evstack *stack)
{
	evstack *stk;
	char buffer[128];

	/* Tell which stack to send info to */
	stk=ev_clone_stack(stack);
	sprintf(buffer,"attach %s",par_stack_to_string(stk));

	termstack->ops->write(termstack,buffer);

	return stk;
}

void usage(char *name)
{ printf("usage : %s [-d] [-g]\n"
         "        -d : attach debug-stack to head\n",name);
  exit(1);
}

int main(int argc, char *argv[])
{


	ggi_event key;
	evstack *termstack[TERMSTACKNUM],*vtswitch,*head,*temp;
	ggi_event *page;
	int i,pagesize;
	char buffer[256];
	FILE *capfile;
	
	int opt_d=0;
	int opt_c=0;

	for(i=1;i<argc;i++)
	{ 
		if (*argv[i]!='-') continue;
		switch(argv[i][1])
		{ case 'd':opt_d=1;break;
		  case 'c':opt_c=1;break;
		  default :usage(argv[0]);break;
		}
	}
	
#ifdef USE_PIPETTY
	init_pipetty();
#endif
	init_graphscroll();
        init_userscroll();

	/* We have to set up the scene ourselves here. */
	
	init_head();
	init_loadkeys();
	init_devecho();
	init_dumb();
	init_xterm();
	init_debug();
	init_capture();
	init_termstack();
	init_vtswitch();
	init_selection();
	init_sound();
	init_joystick();
	init_userkey();
	
	page=ev_evpage_alloc();

	/* Create 10 termstacks */
	for(i=0;i<TERMSTACKNUM;i++)
	{
		evstack *stk;

		termstack[i]=ev_create_stack(ev_lookup_class("termstack"));
		if (termstack[i]==NULL) {
			printk(KERN_ERR "Can't create termstack%d !\n",i);
			return -1;
		}
		/* Populate the termstack */

		create_and_attach_stack(termstack[i],"loadkeys");
#ifdef MEGA_DEBUG
		create_and_attach_stack(termstack[i],"debug");
#endif

		if (i&2) stk=create_and_attach_stack(termstack[i],"xterm");
		else     stk=create_and_attach_stack(termstack[i],"dumbterm");
#ifdef MEGA_DEBUG
		create_and_attach_stack(termstack[i],"debug");
#endif

#ifdef USE_PIPETTY
		if (i>1&&i<4) create_and_attach_stack(termstack[i],"pipetty");
		else          
#endif
			      create_and_attach_stack(termstack[i],"devecho");
#ifdef MEGA_DEBUG
		create_and_attach_stack(termstack[i],"debug");
#endif
		
		clone_and_attach_stack(termstack[i],stk);
#ifdef MEGA_DEBUG
		create_and_attach_stack(termstack[i],"debug");
#endif

		if (i&1) create_and_attach_stack(termstack[i],"graphscroll");
		else     create_and_attach_stack(termstack[i],"userscroll");

#ifdef MEGA_DEBUG
		create_and_attach_stack(termstack[i],"debug");
#endif
		dump_stack(termstack[i],"termstack");
	}

	/* Create the VtSwitcher */
	vtswitch=ev_create_stack(ev_lookup_class("vtswitch"));
	if (termstack==NULL) {
		printk(KERN_ERR "Can't create vtswitch !\n");
		return -1;
	}
	dump_stack(vtswitch,"vtswitch");

	/* attach the termstacks to the VtSwitcher */
	for(i=0;i<TERMSTACKNUM;i++)
	{
		sprintf(buffer,"attach %s %d",par_stack_to_string(termstack[i]),i);
		vtswitch->ops->write(vtswitch,buffer);
	}
	dump_stack(vtswitch,"vtswitch");

	/* Create the Head */
	head=ev_create_stack(ev_lookup_class("head"));
	if (head==NULL) {
		printk(KERN_ERR "Can't create head!\n");
		return -1;
	}

	/* attach the vtswitcher to the head */
	sprintf(buffer,"attach %s",par_stack_to_string(vtswitch));
	head->ops->write(head,buffer);

	/* attach the selection tool to the head */
	temp=ev_create_stack(ev_lookup_class("selection"));
	sprintf(buffer,"attach %s",par_stack_to_string(temp));
	head->ops->write(head,buffer);

	/* attach the joystick */
	temp=ev_create_stack(ev_lookup_class("joystick"));
	sprintf(buffer,"attach %s",par_stack_to_string(temp));
	head->ops->write(head,buffer);

	/* attach the keyboard */
	temp=ev_create_stack(ev_lookup_class("userkey"));
	sprintf(buffer,"attach %s",par_stack_to_string(temp));
	head->ops->write(head,buffer);

	/* attach the soundserver to the head */
	temp=ev_create_stack(ev_lookup_class("sound"));
	sprintf(buffer,"attach %s",par_stack_to_string(temp));
	head->ops->write(head,buffer);
	dump_stack(head,"head");

	/* if asked for, attach the debugger to the head */
	if (opt_d) {
		temp=ev_create_stack(ev_lookup_class("debug"));
		sprintf(buffer,"attach %s",par_stack_to_string(temp));
		head->ops->write(head,buffer);
		dump_stack(head,"head");
	}

	/* if asked for, attach the capture to the head */
	if (opt_c) {
		temp=ev_create_stack(ev_lookup_class("capture"));
		sprintf(buffer,"attach %s",par_stack_to_string(temp));
		head->ops->write(head,buffer);
		dump_stack(head,"head");
	}
	while(1) sleep(1);

	if ((capfile=fopen(argv[1],"r"))!=NULL)
	{
		unsigned char buffer[512];
		int type;
		while(fread(buffer,1,1,capfile)&&*buffer!=0)
		{
			fread(buffer+1,*buffer-1,1,capfile);
			type=(((ggi_event *)buffer)->any.type&=0x7f);
			if (type!=evKeyPress&&type!=evKeyRelease&&type!=evKeyRepeat) continue;
/*			((ggi_event *)buffer)->key.sym=0;*/
			page->size=0;
			pagesize=EVPAGESIZE;
			ev_evpage_add(page,&pagesize,(ggi_event *)buffer);

			head->ops->handler(head,page,&pagesize);
			usleep(50000);
		}
		fclose(capfile);
	}	

	for (i=0;i<10000;i++) {
		/* Start on an empty page */
		page->size=0;
		pagesize=EVPAGESIZE;

		/* Wait for an event, and read it */
		ggiEventPoll(MASTER_VIS,emKeyboard | emPointer,NULL);
		ggiEventRead(MASTER_VIS,(ggi_event *)(&key),emKeyboard | emPointer);

		if (key.any.type==evKeyPress   || key.any.type==evKeyRepeat ||
		    key.any.type==evKeyRelease || key.any.type==evKeyState )
		{
#if 0
			if ((key.key.sym&0xff)=='f') /* 'f' is a hotkey for spit-out 80xA */
			{ 
				DataTtyWriteChar evdata;
				int x;
				evdata.size=80;
				for(x=0;x<80;x++)
					evdata.data[x]='A';
				build_cmd_event(buffer,CmdTtyInWriteChar,&evdata,84);
				ev_evpage_add(page,&pagesize,(ggi_event *)buffer);
			}
			if ((key.key.sym&0xff)=='F') /* 'F' is a hotkey for spit-out 125x80xA */
			{ 	
				DataTtyWriteChar evdata;
				int x;
				evdata.size=80;
				for(x=0;x<80;x++)
					evdata.data[x]='A';
				build_cmd_event(buffer,CmdTtyInWriteChar,&evdata,84);
				for(x=0;x<125;x++)
				{
					ev_evpage_add(page,&pagesize,(ggi_event *)buffer);
					head->ops->handler(head,page,&pagesize);
					page->size=0;
					pagesize=EVPAGESIZE;
				}
			}
#endif
			/* Pretend to not know the keysym... */
			key.key.sym=0;
		}

		/* Add the event to the page */
		ev_evpage_add(page,&pagesize,(ggi_event *)(&key));

		head->ops->handler(head,page,&pagesize);
	}

	ev_evpage_free(page);

	ggiClose(MASTER_VIS);
	ggiExit();

	return 0;
}
