/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>

#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "ggi-dl.h"
#include "Xvisual.h"
#include "../common/X-mode.inc"

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	ggi_mode *mode=VIS_MODE(vis);

	if ( x<0 || x> mode->virt.x-mode->visible.x ||
	     y<0 || y> mode->virt.y-xhook->ysplit )
	     return -EINVAL;

	xhook->xoff=x;
	xhook->yoff=y;
	ggiFlush(vis);
	
	return EOK;
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	struct Xhooks *xhook;
        if (y<0 || y>VIS_MODE(vis)->visible.y) return -1;
        
	xhook=VIS_PRIVATE(vis);

	if ( xhook->yoff > VIS_MODE(vis)->virt.y-y )
	     xhook->yoff = VIS_MODE(vis)->virt.y-y;

	xhook->ysplit=y;
	ggiFlush(vis);
	
	return EOK;
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	char sugname[80];
	
	switch(VIS_MODE(vis)->graphtype)
	{
	  case GT_8BIT : err=8;break;
	  case GT_15BIT: err=15;break;
	  case GT_16BIT: err=16;break;
	  case GT_24BIT: err=24;break;
	  case GT_32BIT: err=32;break;
	  default: return -EINVAL;
	}

	sprintf(sugname, "generic-linear-%d", err);

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	if (_ggiOpenDL(vis,"generic-stubs","") == NULL) {
		fprintf(stderr,"display-X: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	if (_ggiOpenDL(vis,sugname,"") == NULL) {
		fprintf(stderr,"display-X: Can't find an appropriate "
				"library for %s\n", sugname);
		return -1;
	} else {
		DPRINT("Success in loading %s\n",sugname);
	}

	/* Override the color operations for X if we are 8 bit palette. */
	if (XLIB_PRIV(vis)->colormap) { 
		free(XLIB_PRIV(vis)->colormap);XLIB_PRIV(vis)->colormap=NULL;
	}

	if (VIS_MODE(vis)->graphtype==GT_8BIT) {

		int x;
		XLIB_PRIV(vis)->nocols=256;
		XLIB_PRIV(vis)->colormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<XLIB_PRIV(vis)->nocols;x++)
		{
			XLIB_PRIV(vis)->colormap[x].r=((x&0xe0)>>5)*0x2492;
			XLIB_PRIV(vis)->colormap[x].g=((x&0x1c)>>2)*0x2492;
			XLIB_PRIV(vis)->colormap[x].b=((x&0x03)   )*0x5555;
		} 
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
		/* This is just to trigger the initial setup. */
		GGIsetpalvec(vis,0,256,XLIB_PRIV(vis)->colormap);
	}
	/* color */
	vis->opcolor->unmappixel=GGIunmappixel;

	/* X can splitline and pan on a larger bitmap ;-) */
	vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;
        
	return EOK;
}

static int _GGIdoflush(struct Xhooks *xhook)
{
	
	if (xhook->XLibLock++!=0) 
		{xhook->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/

	if (xhook->cmap && xhook->cmap_first<xhook->cmap_last) {

		int x;
		XColor xcol;
	       	for(x=xhook->cmap_first;x<xhook->cmap_last;x++)
       		{
       			xcol.red  =xhook->colormap[x].r;
	       		xcol.green=xhook->colormap[x].g;
       			xcol.blue =xhook->colormap[x].b;
       			xcol.pixel=x;
	       		xcol.flags= DoRed | DoGreen | DoBlue ;
       			XStoreColor(xhook->display, xhook->cmap,&xcol);
	       	}
		xhook->cmap_first=256;
		xhook->cmap_last=0;
		XSetWindowColormap(xhook->display, xhook->window, xhook->cmap);
	}

#ifndef NO_SHM
	if (xhook->have_shm)
	{
		XShmPutImage(xhook->display, xhook->window, xhook->gc,
			  xhook->ximage, 
			  xhook->xoff, xhook->yoff, /* panning ! */
			  0, 0,
			  xhook->viswidth,
			  xhook->ysplit, 0);

		if (xhook->ysplit<xhook->visheight)
			XShmPutImage(xhook->display, xhook->window, xhook->gc,
				  xhook->ximage, 
				  0, 0,
				  0, xhook->ysplit,
				  xhook->viswidth,
				  xhook->visheight-xhook->ysplit, 0);
	} else 
#endif
		{

		XPutImage(xhook->display, xhook->window, xhook->gc,
			  xhook->ximage, 
			  xhook->xoff, xhook->yoff, /* panning ! */
			  0, 0,
			  xhook->viswidth, 
			  xhook->ysplit);

		if (xhook->ysplit<xhook->visheight)
			XPutImage(xhook->display, xhook->window, xhook->gc,
				  xhook->ximage, 
				  0, 0,
				  0, xhook->ysplit, /* panning ! */
				  xhook->viswidth, 
				  xhook->visheight-xhook->ysplit);
	}

	XSync(xhook->display,0);
	xhook->XLibLock--;	/* FIXME Spinlocks ... */

	return EOK;
}

static struct Xhooks *xhk;
#include "../common/mansync.inc"

#ifndef NO_SHM
/* Hack to fall back when shm is not working. */
static int	shmerror;
static int	(*oldshmerrorhandler) (Display *, XErrorEvent *);

static int shmerrorhandler (Display * disp, XErrorEvent * event)
{
	if (event->error_code == BadAccess)
		shmerror=1;
	else
		(*oldshmerrorhandler) (disp, event);
	return 0;
}
#endif

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct Xhooks *xhook;
	XSizeHints hint;
	unsigned int fg, bg;
	XVisualInfo vinfo;
	int rc;
	
	if ((rc=GGIcheckmode(vis,tm))) return rc;

	xhook=VIS_PRIVATE(vis);

	if (XLIB_PRIV(vis)->XLibLock++!=0) 
		{XLIB_PRIV(vis)->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/
	if (!(xhook->isasync))
		_mansync_ignore(vis);
	
	_GGIbasiccheck(vis,tm,&vinfo);
	
	xhook->xoff=xhook->yoff=0;
	xhook->ysplit=tm->visible.y;
	xhook->viswidth=tm->visible.x;
	xhook->visheight=tm->visible.y;

	/* Fill in hint structure. Minimal window until we know better. */
	hint.x      = hint.y      = 0;
	hint.width  = tm->visible.x;
	hint.height = tm->visible.y;
	hint.flags  = PSize | PMinSize | PMaxSize;
	hint.min_width = tm->visible.x;
	hint.min_height = tm->visible.y;
	hint.max_width = tm->visible.x;
	hint.max_height = tm->visible.y;
  
	/* Get some colors */
	bg = WhitePixel (xhook->display, xhook->screen);
	fg = BlackPixel (xhook->display, xhook->screen);
	DPRINT("X-lib has colors.\n");
  
	if (xhook->window) 
		XDestroyWindow(xhook->display,xhook->window);

	if (xhook->ximage)
		XDestroyImage(xhook->ximage);

#ifndef NO_SHM
	if (xhook->shminfo.shmid!=-1) {	
		XShmDetach(xhook->display, &(xhook->shminfo));
		DPRINT("Detach\n");
		if (VIS_FB_LINEAR(vis)!=NULL) {
			shmdt(VIS_FB_LINEAR(vis));
		}
	}
#endif
	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}


	xhook->window = XCreateSimpleWindow (xhook->display,
				 RootWindow (xhook->display,xhook->screen),
				 hint.x, hint.y,
				 hint.width, hint.height,
				 2, fg, bg);
  
	DPRINT("X-lib has window.\n");

	XSelectInput(xhook->display, xhook->window,
	             KeyPressMask|KeyReleaseMask|
	             ButtonPressMask|ButtonReleaseMask|
	             PointerMotionMask);

	/* Tell other applications about this window */
	XSetStandardProperties (xhook->display, xhook->window, "GGI-on-X", "GGI-on-X", None, NULL, 0, &hint);

	if (!xhook->gc)
	  	xhook->gc = XCreateGC(xhook->display, xhook->window, 0, 0);
	DPRINT("X-lib GC allocated.\n");

#ifndef NO_SHM
	if (xhook->have_shm)
	{
		DPRINT("X: mode setting\n");
		xhook->ximage = XShmCreateImage(
			xhook->display, vinfo.visual, vinfo.depth,
			ZPixmap, 
			0, 
			&(xhook->shminfo),
			tm->virt.x,tm->virt.y);

#ifdef LITTLE_ENDIAN_ARCHITECTURE
		xhook->ximage->byte_order = LSBFirst;
		xhook->ximage->bitmap_bit_order = LSBFirst;
#else
		xhook->ximage->byte_order = MSBFirst;
		xhook->ximage->bitmap_bit_order = MSBFirst;
#endif

		xhook->shminfo.shmid = 
			shmget(IPC_PRIVATE, (unsigned int) 
				tm->virt.x*tm->virt.y*vinfo.depth,
				IPC_CREAT | 0777);

		VIS_FB_LINEAR(vis)=xhook->shminfo.shmaddr=xhook->ximage->data = 
			(char *) shmat(xhook->shminfo.shmid,0,0);
		xhook->shminfo.readOnly=False;
	
		DPRINT("X: shmat success at %p.\n",VIS_FB_LINEAR(vis));

		/* Argh ... X sometimes says it can XShm but it can't ... */
		shmerror = 0;
		oldshmerrorhandler = XSetErrorHandler(shmerrorhandler);
		XShmAttach(xhook->display, &(xhook->shminfo));

		XSync(xhook->display,0);
		XSetErrorHandler(oldshmerrorhandler);

		if (shmerror) {
			if (VIS_FB_LINEAR(vis)!=NULL) {
				shmdt(VIS_FB_LINEAR(vis));
				VIS_FB_LINEAR(vis)=NULL;
			}
			if (xhook->ximage) {
				XDestroyImage(xhook->ximage);
				xhook->ximage=NULL;
			}
			xhook->shminfo.shmid = -1;
			xhook->have_shm = 0;
			DPRINT("X: XShmAttach failed - trying without shm.\n");
		} else {
			shmctl(xhook->shminfo.shmid, IPC_RMID, 0);
			DPRINT("X: ShmAllDone.\n");
		}
	}

	if (! xhook->have_shm)
#endif
	{
		VIS_FB_LINEAR(vis)= 
			(char *) malloc(tm->virt.x*tm->virt.y*vinfo.depth);
		DPRINT("X: Buffer allocated %p.\n",VIS_FB_LINEAR(vis));

		xhook->ximage = XCreateImage(
			xhook->display, vinfo.visual, vinfo.depth,
			ZPixmap, /* format */
			0, 	 /* offset */
			VIS_FB_LINEAR(vis),	/* data */
			tm->virt.x,tm->virt.y, /* size */
			8,	/* bitmap_pad*/
			0 
			);

		DPRINT("X: Image allocated.\n");
	}

	XResizeWindow(xhook->display, xhook->window, tm->visible.x, tm->visible.y);
	DPRINT("X: Resize.\n");

	DPRINT("About to map.\n");
	/* Map window. */
	XSelectInput (xhook->display, xhook->window, ExposureMask);
	XMapRaised(xhook->display, xhook->window);
	XFlush(xhook->display);
		/* Wait for window to become mapped */
	{
		XEvent        event;
		XNextEvent (xhook->display, &event);
	}
	DPRINT("X: Window Mapped.\n");

		/* Select input events to listen for */
	XSelectInput(xhook->display, xhook->window, 
				 KeyPressMask|KeyReleaseMask|
				 ButtonPressMask|ButtonReleaseMask|
				 PointerMotionMask);

	if (xhook->cmap)
		XFreeColormap(xhook->display,xhook->cmap);

	if (tm->graphtype==GT_8BIT) {
		xhook->cmap = XCreateColormap(xhook->display, xhook->window, 
									  vinfo.visual, 
									  AllocAll);
		 
		xhook->cmap_first=256;
		xhook->cmap_last=0;
		 
		DPRINT("X-lib colormap allocated %x.\n",xhook->cmap);
	} else {
		xhook->cmap = XCreateColormap(xhook->display, xhook->window, 
									  vinfo.visual, 
									  AllocNone);
	}

	_GGIdoflush(xhk=xhook);
	
	DPRINT("X: Flush.\n");
	
	if (!(xhook->isasync))
		_mansync_cont(vis);

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	xhook->XLibLock--;	/* FIXME Spinlocks ... */

	return _GGIdomode(vis);
}
