/* 
   Graphics library for GGI. Copying Boxes.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   
   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ggi-dl.h"
#include "Xvisual.h"

int GGIdrawbox(ggi_visual *vis, int x, int y, int w, int h)
{
	XFillRectangle(XLIB_PRIV(vis)->display,
				   XLIB_PRIV(vis)->window,
				   XLIB_PRIV(vis)->gc,
				   x, y, w, h);

	XLIB_DOSYNC(vis);
	return EOK;
}

int GGIputbox(ggi_visual *vis, int x, int y, int w, int h, void *data)
{
	XImage *ximg;
	
	ximg=XCreateImage(XLIB_PRIV(vis)->display, 
					  DefaultVisual(XLIB_PRIV(vis)->display, 0),
					  VIS_FB_BPP(vis), ZPixmap, 0,
					  data, w, h, 8, 0);
	XPutImage(XLIB_PRIV(vis)->display, 
			  XLIB_PRIV(vis)->window,
			  XLIB_PRIV(vis)->gc,
			  ximg, 0, 0, x, y, w, h);
	XFree(ximg);

	XLIB_DOSYNC(vis);
	return EOK;
}

int GGIcopybox(ggi_visual *vis, int x, int y, 
			   int w, int h, int nx, int ny)
{
	XCopyArea(XLIB_PRIV(vis)->display, 
			  XLIB_PRIV(vis)->window,
			  XLIB_PRIV(vis)->window,
			  XLIB_PRIV(vis)->gc,
			  x, y, w, h, nx, ny);

	XLIB_DOSYNC(vis);
	return EOK;
}

int GGIgetbox(ggi_visual *vis, int x, int y, int w, int h, void *data)
{
	XImage *ximg;

	ximg = XGetImage(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->window, x, y,
			  w, h, AllPlanes, ZPixmap);
	memcpy(data, ximg->data, w*h*VIS_FB_BPP(vis)/8);
	XFree(ximg);

	XLIB_DOSYNC(vis);
	return EOK;
}
