/* ---------------------------------------------------------------------------
** LibGGI GLIDE target - Mode handling
** ---------------------------------------------------------------------------
** Copyright (C) 1997, Jon Taylor <taylorj@ecs.csus.edu>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>

#include "ggi-dl.h"
#include "internal.h"
#include "glidevisual.h"

#include <ggi/ggi_commands.h>

extern int        _GGIbpc;
extern int        _GGInocols;
extern ggi_color *_GGIcolormap;

int GGIsetorigin(ggi_visual *vis, int x, int y)
{
	
	return E(DRIVER,NOSUP_ALWAYS_CANT);
}

int GGIsetsplitline(ggi_visual *vis, int y)
{
	
	return E(DRIVER,NOSUP_ALWAYS_CANT);
}

int GGIdoflush(struct glide_hooks *glidehook)
{

	DPRINT("buffer = %d, x_start = %d, y_start = %d,",
	       glidehook->buffer, glidehook->x_start, glidehook->y_start);
	DPRINT("src_format = %d, width = %d, height = %d,",
	       glidehook->src_format, glidehook->width, glidehook->height);
	DPRINT("stride_in_bytes = %d, data = %p",
	       glidehook->stride_in_bytes, glidehook->data);
	
	grLfbWriteRegion(glidehook->buffer, glidehook->x_start, 
			 glidehook->y_start, glidehook->src_format,
			 glidehook->width, glidehook->height, 
			 glidehook->stride_in_bytes, glidehook->data);
	
	return EOK;
}

static const char *_GGIgetmmap(ggi_visual *vis, ggi_mode *mode)
{
	int size;
	const char *cp = NULL;
	
	cp = "generic-linear-16";
	
	if(VIS_FB_LINEAR(vis) != NULL)
	{
		
		free(VIS_FB_LINEAR(vis));
	}
	
	size = _ggiSetupMode(vis);
	if (size <= 0)
	{
		
		return cp;
	}
	
	VIS_FB_LINEAR_SIZE(vis) = size;
	VIS_FB_LINEAR(vis) = malloc(size);
	
	if (VIS_FB_LINEAR(vis) == NULL)
	{
	
		VIS_FB_LINEAR_SIZE(vis) = 0;
		DPRINT("Failed to mmap a %dK area", size / 1024);
		return NULL;
	}
	
	return cp;
}


static int _GGIdomode(ggi_visual *vis, ggi_mode *mode)
{
	int err;
	const char *name;
	
	_ggiZapMode(vis, ~GGI_DL_OPDISPLAY);
	
	name = _GGIgetmmap(vis, mode);
	if (name == NULL)
	{
		
		return -1;
	}
	
	err = (_ggiOpenDL(vis, "generic-stubs", "") == NULL);
	if (err) 
	{
		
		DPRINT("display-glide: Can't open the generic-stubs library");
	}
	
	err = (_ggiOpenDL(vis, name, "") == NULL);
	if (err)
	{
		
		DPRINT("display-glide: Can't open the generic-linear-16 library");
	}
		
	return EOK;
}

static struct glide_hooks *glhk;

void _doflush_ALRM(int unused)
{
	
	struct itimerval tval = {{0, 0},{0, 0}};
	signal(SIGALRM, _doflush_ALRM);
	GGIdoflush(glhk);
	setitimer(ITIMER_REAL, &tval, NULL);
}

int GGIflush(ggi_visual *vis)
{

	if (VIS_FLAGS(vis) & GGIFLAG_ASYNC)
	{
	
		struct itimerval tval = {{0, 0},{0, 0}};
		signal(SIGALRM, _doflush_ALRM);
		setitimer(ITIMER_REAL, &tval, NULL);
	}
	else
	{
	
		struct itimerval tval = {{0, 0},{0, 50000}};
		signal(SIGALRM, _doflush_ALRM);
		setitimer(ITIMER_REAL, &tval, NULL);
	}
	
	return GGIdoflush((struct glide_hooks *)VIS_PRIVATE(vis));
}

		

int GGIsetmode(ggi_visual *vis, ggi_mode *mode)
{
	struct glide_hooks *glidehook;

	GrScreenResolution_t resolution;
	
	if(GGIcheckmode(vis, mode))
		return -1;

	glidehook = VIS_PRIVATE(vis);

	if (VIS_FB_LINEAR(vis) != NULL) 
	{
	      
		VIS_FB_LINEAR(vis) = NULL;
		VIS_FB_LINEAR_SIZE(vis) = 0;
	}
       
	glidehook->buffer = GR_BUFFER_FRONTBUFFER;
	glidehook->x_start = 0;
	glidehook->y_start = 0;
	glidehook->width = mode->visible.x;
	glidehook->height = mode->visible.y;
	
	switch(mode->graphtype)
	{
	
		case GT_15BIT:
		glidehook->src_format = GR_LFB_SRC_FMT_555;
		glidehook->stride_in_bytes = mode->visible.x * 2;
		break;
		
		case GT_16BIT:
		glidehook->src_format = GR_LFB_SRC_FMT_565;
		glidehook->stride_in_bytes = mode->visible.x * 2;
		break;

		case GT_24BIT:
		glidehook->src_format = GR_LFB_SRC_FMT_8888;
		glidehook->stride_in_bytes = mode->visible.x * 2;
		break;

		case GT_32BIT:
		glidehook->src_format = GR_LFB_SRC_FMT_8888;
		glidehook->stride_in_bytes = mode->visible.x * 2;
		break;
		default:
		DPRINT("Illegal mode!");
		return GGI_DL_ERROR;
	}
	
	/* FIXME: Allocate the proper size for the bit depth,
	** not always 32
	*/
	VIS_FB_LINEAR(vis) = (char *)malloc(mode->virt.x * mode->virt.y * 32);
	glidehook->data = VIS_FB_LINEAR(vis);
	DPRINT("GLIDE: Buffer allocated %p", VIS_FB_LINEAR(vis));
	
	if (mode->visible.x == 640 && mode->visible.y == 480)
	{
		
		resolution = GR_RESOLUTION_640x480;
	}
	else if (mode->visible.x == 800 && mode->visible.y == 600)
	{
		
		resolution = GR_RESOLUTION_800x600;
	}
	else if (mode->visible.x == 512 && mode->visible.y == 384)
	{
		
		resolution = GR_RESOLUTION_512x384;
	}
	else 
	{
		DPRINT("Invalid Glide resolution (%dx%d)", mode->visible.x,
		       mode->visible.y);
		return GGI_DL_ERROR;
	}

	
	grSstWinOpen(0, resolution, GR_REFRESH_60Hz,
		     GR_COLORFORMAT_ABGR, GR_ORIGIN_UPPER_LEFT, 2, 1);
	grBufferClear(0xff0000, 0, GR_WDEPTHVALUE_FARTHEST);
	grBufferSwap(1);
	grClipWindow(0,0, (FxU32) mode->visible.x, (FxU32) mode->visible.y);

	GGIdoflush(glhk = glidehook);
	signal(SIGALRM, _doflush_ALRM);
	alarm(1);
	
	memcpy(VIS_MODE(vis), mode, sizeof(ggi_mode));
	return _GGIdomode(vis, mode);
}

/*
** Check any mode (text/graphics)
*/
int GGIcheckmode(ggi_visual *vis, ggi_mode *tm)
{
	int err = 0;

	if (vis == NULL)
	{
		DPRINT("Visual==NULL\n");
		return -1;
	}

	/* FIXME: There are some hardcoded values here; probably it should
	   check memory available, etc., as well as being smarter when
	   resetting GGI_AUTO */

	if(tm->visible.x==GGI_AUTO) tm->visible.x = 512;
	if(tm->visible.y==GGI_AUTO) tm->visible.y = 384;

	if(!(	(tm->visible.x == 512 && tm->visible.y == 384) ||
		(tm->visible.x == 640 && tm->visible.y == 480) ||
		(tm->visible.x == 800 && tm->visible.y == 600)))
	{
		tm->visible.x = 512;
		tm->visible.y = 384;
		err = -1;
	}

	if(tm->virt.x < tm->visible.x) {
		tm->virt.x = tm->visible.x;
		err = -1;
	}

	if(tm->virt.y < tm->visible.y) {
		tm->virt.y = tm->visible.y;
		err = -1;
	}

        if(     (tm->dpp.x!=1 && tm->dpp.x!=GGI_AUTO) ||
                (tm->dpp.y!=1 && tm->dpp.y!=GGI_AUTO))
                { err = -1; }
        tm->dpp.x = tm->dpp.y = 1;

	switch(tm->graphmode) {
		case GT_15BIT:
		case GT_16BIT:
		case GT_24BIT:
		case GT_32BIT: break;
		case GGI_AUTO: tm->graphmode=GT_16BIT; break;
		default: tm->graphmode=GT_16BIT; err = -1;
	}

        return err;
}

/*
** Get the current mode
*/
int GGIgetmode(ggi_visual *vis, ggi_mode *tm)
{
	
	DPRINT("In GGIgetmode(%p,%p)", vis, tm);
	
	if (vis == NULL)
	{
		
		return -1;
	}
	
	/* We assume the mode in the visual to be OK 
	*/
	memcpy(tm, VIS_MODE(vis), sizeof(ggi_mode));

	return 0;
}
