/* 
   Graphics library for GGI. Events for SVGAlib.

   Copyright (C) 1997 Steve Cheng <elmert@ipoline.com>
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <string.h>

#include "ggi-dl.h"

#include "vgavisual.h"
#include "../common/evqueue.inc"

extern char *_GGIkeystate;
extern ggi_event_buffer *_GGIevbuf;

void _GGIkeyboardhandler(int scancode, int press)
{
	ggi_event ggievent;

	ggievent.any.focus=0;
	ggievent.any.device=0;
	ggievent.any.time=0;    /* FIXME ! */
	ggievent.any.size=sizeof(ggi_key_event);

	if(press) {
		if(_GGIkeystate[scancode]) {
			ggievent.any.type=evKeyRepeat;
			ggievent.key.code=scancode;
		}

		ggievent.any.type=evKeyPress;
		ggievent.key.code = scancode;
		_GGIkeystate[scancode] = 1;
	}
	else {
		ggievent.any.type=evKeyRelease;
		ggievent.key.code = scancode;
		_GGIkeystate[scancode] = 0;
	}	

	/* This mapping isn't correct. key.sym is supposed to
	   be a 'cooked' symbol combining the various modifiers,
	   which this handler doesn't do. However, a new field
	   called BaseSym will be added in the future which is
	   a portable version of key.code. */
	switch(scancode) {
		case SCANCODE_ESCAPE: ggievent.key.sym=27 ; break;

		case SCANCODE_1: ggievent.key.sym='1' ; break;
		case SCANCODE_2: ggievent.key.sym='2' ; break;
		case SCANCODE_3: ggievent.key.sym='3' ; break;
		case SCANCODE_4: ggievent.key.sym='4' ; break;
		case SCANCODE_5: ggievent.key.sym='5' ; break;
		case SCANCODE_6: ggievent.key.sym='6' ; break;
		case SCANCODE_7: ggievent.key.sym='7' ; break;
		case SCANCODE_8: ggievent.key.sym='8' ; break;
		case SCANCODE_9: ggievent.key.sym='9' ; break;
		case SCANCODE_0: ggievent.key.sym='0' ; break;

		case SCANCODE_MINUS: ggievent.key.sym='-' ; break;
		case SCANCODE_EQUAL: ggievent.key.sym='=' ; break;
		case SCANCODE_BACKSPACE: ggievent.key.sym=8 ; break;
		case SCANCODE_TAB: ggievent.key.sym=9 ; break;

		case SCANCODE_Q: ggievent.key.sym='q' ; break;
		case SCANCODE_W: ggievent.key.sym='w' ; break;
		case SCANCODE_E: ggievent.key.sym='e' ; break;
		case SCANCODE_R: ggievent.key.sym='r' ; break;
		case SCANCODE_T: ggievent.key.sym='t' ; break;
		case SCANCODE_Y: ggievent.key.sym='y' ; break;
		case SCANCODE_U: ggievent.key.sym='u' ; break;
		case SCANCODE_I: ggievent.key.sym='i' ; break;
		case SCANCODE_O: ggievent.key.sym='o' ; break;
		case SCANCODE_P: ggievent.key.sym='p' ; break;

		case SCANCODE_BRACKET_LEFT: ggievent.key.sym='[' ; break;
		case SCANCODE_BRACKET_RIGHT: ggievent.key.sym=']' ; break;
		case SCANCODE_ENTER: ggievent.key.sym=K_ENTER ; break;
		case SCANCODE_LEFTCONTROL: ggievent.key.sym=K_NORMAL_CTRLL ; break;

		case SCANCODE_A: ggievent.key.sym='a' ; break;
		case SCANCODE_S: ggievent.key.sym='s' ; break;
		case SCANCODE_D: ggievent.key.sym='d' ; break;
		case SCANCODE_F: ggievent.key.sym='f' ; break;
		case SCANCODE_G: ggievent.key.sym='g' ; break;
		case SCANCODE_H: ggievent.key.sym='h' ; break;
		case SCANCODE_J: ggievent.key.sym='j' ; break;
		case SCANCODE_K: ggievent.key.sym='k' ; break;
		case SCANCODE_L: ggievent.key.sym='l' ; break;

		case SCANCODE_SEMICOLON: ggievent.key.sym=';' ; break;
		case SCANCODE_APOSTROPHE: ggievent.key.sym='\'' ; break;
		case SCANCODE_GRAVE: ggievent.key.sym='`' ; break;
		case SCANCODE_LEFTSHIFT: ggievent.key.sym=K_NORMAL_SHIFTL ; break;
		case SCANCODE_BACKSLASH: ggievent.key.sym='\\' ; break;

		case SCANCODE_Z: ggievent.key.sym='z' ; break;
		case SCANCODE_X: ggievent.key.sym='x' ; break;
		case SCANCODE_C: ggievent.key.sym='c' ; break;
		case SCANCODE_V: ggievent.key.sym='v' ; break;
		case SCANCODE_B: ggievent.key.sym='b' ; break;
		case SCANCODE_N: ggievent.key.sym='n' ; break;
		case SCANCODE_M: ggievent.key.sym='m' ; break;

		case SCANCODE_COMMA: ggievent.key.sym=',' ; break;
		case SCANCODE_PERIOD: ggievent.key.sym='.' ; break;
		case SCANCODE_SLASH: ggievent.key.sym='/' ; break;
		case SCANCODE_RIGHTSHIFT: ggievent.key.sym=K_NORMAL_SHIFTR ; break;
		case SCANCODE_KEYPADMULTIPLY: ggievent.key.sym=K_PSTAR ; break;
		case SCANCODE_LEFTALT: ggievent.key.sym=K_NORMAL_ALT ; break;
		case SCANCODE_SPACE: ggievent.key.sym=' ' ; break;

/* ??? */		case SCANCODE_CAPSLOCK: ggievent.key.sym=K_CAPS ; break;

		case SCANCODE_F1: ggievent.key.sym=K_F1 ; break;
		case SCANCODE_F2: ggievent.key.sym=K_F2 ; break;
		case SCANCODE_F3: ggievent.key.sym=K_F3 ; break;
		case SCANCODE_F4: ggievent.key.sym=K_F4 ; break;
		case SCANCODE_F5: ggievent.key.sym=K_F5 ; break;
		case SCANCODE_F6: ggievent.key.sym=K_F6 ; break;
		case SCANCODE_F7: ggievent.key.sym=K_F7 ; break;
		case SCANCODE_F8: ggievent.key.sym=K_F8 ; break;
		case SCANCODE_F9: ggievent.key.sym=K_F9 ; break;
		case SCANCODE_F10: ggievent.key.sym=K_F10 ; break;

		case SCANCODE_NUMLOCK: ggievent.key.sym=K_BARENUMLOCK ; break;
/* ??? */		case SCANCODE_SCROLLLOCK: ggievent.key.sym=K_HOLD ; break;

		case SCANCODE_KEYPAD7: ggievent.key.sym=K_P7 ; break;
		case SCANCODE_KEYPAD8: ggievent.key.sym=K_P8 ; break;
		case SCANCODE_KEYPAD9: ggievent.key.sym=K_P9 ; break;
		case SCANCODE_KEYPADMINUS: ggievent.key.sym=K_PMINUS ; break;
		case SCANCODE_KEYPAD4: ggievent.key.sym=K_P4 ; break;
		case SCANCODE_KEYPAD5: ggievent.key.sym=K_P5 ; break;
		case SCANCODE_KEYPAD6: ggievent.key.sym=K_P6 ; break;
		case SCANCODE_KEYPADPLUS: ggievent.key.sym=K_PPLUS ; break;
		case SCANCODE_KEYPAD1: ggievent.key.sym=K_P1 ; break;
		case SCANCODE_KEYPAD2: ggievent.key.sym=K_P2 ; break;
		case SCANCODE_KEYPAD3: ggievent.key.sym=K_P3 ; break;
		case SCANCODE_KEYPAD0: ggievent.key.sym=K_P0 ; break;
		case SCANCODE_KEYPADPERIOD: ggievent.key.sym=K_PDOT ; break;
		case SCANCODE_LESS: ggievent.key.sym=K_PPLUSMINUS ; break;

		case SCANCODE_F11: ggievent.key.sym=K_F11 ; break;
		case SCANCODE_F12: ggievent.key.sym=K_F12 ; break;

		case SCANCODE_KEYPADENTER: ggievent.key.sym=K_PENTER ; break;
		case SCANCODE_CONTROL: ggievent.key.sym=K_NORMAL_CTRL ; break;
		case SCANCODE_KEYPADDIVIDE: ggievent.key.sym=K_PSLASH ; break;
		case SCANCODE_PRINTSCREEN: ggievent.key.sym=K_SAK ; break;
		case SCANCODE_RIGHTALT: ggievent.key.sym=K_NORMAL_SHIFTR ; break;
		case SCANCODE_BREAK: ggievent.key.sym=K_BREAK; break;
		case SCANCODE_BREAK_ALTERNATIVE: ggievent.key.sym=K_PAUSE ; break;

/* ??? */	case SCANCODE_HOME: break;
		case SCANCODE_CURSORBLOCKUP: ggievent.key.sym=K_UP ; break;
		case SCANCODE_PAGEUP: ggievent.key.sym=K_PGUP ; break;
		case SCANCODE_CURSORBLOCKLEFT: ggievent.key.sym=K_LEFT ; break;
		case SCANCODE_CURSORBLOCKRIGHT: ggievent.key.sym=K_RIGHT ; break;
/* ??? */       case SCANCODE_END: break;
		case SCANCODE_CURSORBLOCKDOWN: ggievent.key.sym=K_DOWN ; break;
		case SCANCODE_PAGEDOWN: ggievent.key.sym=K_PGDN ; break;
		case SCANCODE_INSERT: ggievent.key.sym=K_INSERT ; break;
		case SCANCODE_REMOVE: ggievent.key.sym=K_REMOVE ; break;
	}

	ggievent.key.sym = U(ggievent.key.sym);

	if(ggiEventPut(_GGIevbuf, &ggievent))
		DPRINT("kbd_handler: couldn't put event!\n");
}	

void _GGImousehandler(int button, int dx, int dy, int dz,
	int drx, int dry, int drz)
{
	static int last_button = 0;
	int button_change;
	ggi_event ggievent;

	ggievent.any.focus=0;
	ggievent.any.device=0;
	ggievent.any.time=0;    /* FIXME ! */

	/* Mouse movement */
	if(dx || dy || dz || drx || dry || drz) {
		ggievent.any.type=evPtrRelative;
		ggievent.any.size=sizeof(ggi_pmove_event);
		ggievent.pmove.x=dx;
		ggievent.pmove.y=dy;
		ggievent.pmove.wheel=dz;

		DPRINT("mouse move: dx: %d, dy: %d, dz: %d, drx: %d, dry: %d, drz: %d\n",
			dx, dy, dz, drx, dry, drz);
		
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	
	/* Buttons changed? */
	button_change = button ^ last_button;
	last_button = button;

	ggievent.any.size=sizeof(ggi_pbutton_event);

	if(button_change & MOUSE_LEFTBUTTON) {
		ggievent.pbutton.button = 1;
		if(button & MOUSE_LEFTBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_RIGHTBUTTON) {
		ggievent.pbutton.button = 2;
		if(button & MOUSE_RIGHTBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_MIDDLEBUTTON) {
		ggievent.pbutton.button = 4;
		if(button & MOUSE_MIDDLEBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_FOURTHBUTTON) {
		ggievent.pbutton.button = 8;
		if(button & MOUSE_FOURTHBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_FIFTHBUTTON) {
		ggievent.pbutton.button = 16;
		if(button & MOUSE_FIFTHBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_SIXTHBUTTON) {
		ggievent.pbutton.button = 32;
		if(button & MOUSE_SIXTHBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
	if(button_change & MOUSE_RESETBUTTON) {
		ggievent.pbutton.button = 64;
		if(button & MOUSE_RESETBUTTON) {
			ggievent.pbutton.state = 1;
			ggievent.any.type = evPtrButtonPress;
		} else {
			ggievent.pbutton.state = 0;
			ggievent.any.type = evPtrButtonRelease;
		}
		if(ggiEventPut(_GGIevbuf, &ggievent))
			DPRINT("mouse_handler: couldn't put event!\n");
	}
}

/************************** GGI Functions ***************************/
/* Event Handling */
ggi_event_mask GGIeventpoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{
	fd_set fds;
	int err;
	struct timeval *timeout,t_zero={0,0};
	ggi_event_mask evmask;
	ggi_event ggievent;

DPRINT("GGIeventpoll(%p,0x%.8x",vis,mask);
if (t==NULL) {
	DPRINT(",NULL)\n");
} else {
	DPRINT(",{%d,%d})\n",t->tv_sec,t->tv_usec);
}

	evmask=_ggiEvQueueSeen(vis,mask);
	if (evmask!=0)
		return evmask;

	if (VIS_SELECT_FD(vis) < 0)
		return 0;

	timeout = &t_zero;
	do {

		/* SVGAlib bug compability!!! input fds can't be null pointer */
		FD_ZERO(&fds);
		err=vga_waitevent(VGA_KEYEVENT | VGA_MOUSEEVENT,
			&fds,NULL,NULL,timeout);

		if(err<0)
			return 0;

		timeout=t;

		while(ggiEventAvailable(_GGIevbuf)) {
			ggiEventCopy(_GGIevbuf, &ggievent);
			ggiEventSkip(_GGIevbuf);
		
			switch(ggievent.any.type) {
				case evKeyPress:
					DPRINT("evKeyPress: scancode %d, sym %d\n",
						ggievent.key.code, ggievent.key.sym);
					break;
				case evKeyRelease:
					DPRINT("evKeyRelease: scancode %d, sym %d\n",
						ggievent.key.code, ggievent.key.sym);
					break;
				case evPtrRelative:
					DPRINT("evPtrRelative: dx %d, dy %d\n",
						ggievent.pmove.x, ggievent.pmove.y);
					break;
				case evPtrButtonPress:
				case evPtrButtonRelease:
					DPRINT("evPtrButton: button %d, state %d\n",
						ggievent.pbutton.button, ggievent.pbutton.state);
			}		

			_ggiEvQueueAdd(vis,&ggievent);
			evmask = _ggiEvQueueSeen(vis,mask);
		}
	} while (evmask==0 && t==NULL);

	return evmask;
}

int GGIeventread(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{
	/* Block if we don't have anything queued... */
	GGIeventpoll(vis,mask,NULL);
	return _ggiEvQueueRelease(vis,ev,mask);
}
