/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   Copyright (C) 1998 Steve Cheng <elmert@ipoline.com>
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>

#include "ggi-dl.h"
#include "vgavisual.h"

int GGIwaitraypos(ggi_visual *vis,sint32 *x,sint32 *y)
{
	if (*y == GGIRP_SYNC && *x == GGIRP_DONTCARE) {
		vga_waitretrace();
		return 0;
	}
	*x = GGIRP_DONTCARE;
	*y = GGIRP_SYNC;

	return 1;
}

#if defined(__linux__) && defined (WANT_DANGEROUS_SPLITLINE)
#include <asm/io.h>
int GGIsetsplitline(ggi_visual *vis,int y)
{
#define CRT_LINECOMPARE      0x18
#define CR07_LINECOMPARE_B8  0x10
#define CR09_LINECOMPARE_B9  0x40
#define outbCRT(val,reg) outw(((val) << 8) | (reg), CRTI)
	register uint8 foo;
	unsigned short CRTI,CRTD;

	if (y<0 || y>VIS_MODE(vis)->visible.y) return -1;
	if (VIS_MODE(vis)->visible.y<0x100) y <<= 1;

	CRTI=((struct svga_priv*)VIS_PRIVATE(vis))->CRTI;
	CRTD=((struct svga_priv*)VIS_PRIVATE(vis))->CRTD;

	outbCRT(y & 0xFF, CRT_LINECOMPARE);

	outb(0x07, CRTI);

	foo = (y & 0x0100)
		? inb(CRTD) | CR07_LINECOMPARE_B8
		: inb(CRTD) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	outb(0x09, CRTI);
	foo = (y & 0x0200)
		? inb(CRTD) | CR09_LINECOMPARE_B9
		: inb(CRTD) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);
	
	return EOK;
}
#endif

/*int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	if (x != 0 || y<0 || y> VIS_MODE(vis)->virt.y )
		return -EINVAL;

	vga_setdisplaystart(y * (VIS_FB_BPP(vis)*VIS_MODE(vis)->virt.x)/8
						+ (x * VIS_FB_BPP(vis))/8);
	
	return EOK;
}*/

int GGIflush(ggi_visual *vis)
{
	return EOK;
}
	
int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	char modestr[64], *colors;
	int modenum, depth;
	vga_modeinfo *modeinfo;
	int err;
	struct svga_priv *visinfo;

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}

	if(GGIcheckmode(vis, tm))
		return -1;
	
	/* See GGIcheckmode() for details... */
	colors = "0"; depth = 0;
	switch(tm->graphtype) {
	case GT_1BIT : colors = "2"; depth = 1; break;
	case GT_4BIT : colors = "16"; depth = 4; break;
	case GT_8BIT : colors = "256"; depth = 8; break;
	case GT_15BIT: colors = "32K"; depth = 16; break;
	case GT_16BIT: colors = "64K"; depth = 16; break;
	case GT_24BIT : colors = "16M"; depth = 24; break;
	case GT_32BIT : colors = "16M4"; depth = 32; break;
	default: return -EINVAL;
	}

	/* Form a SVGAlib mode number */
	sprintf(modestr, "G%dx%dx%s", tm->visible.x, tm->visible.y, colors);
 	modenum = vga_getmodenumber(modestr);
	DPRINT("Setting SVGAlib mode %d: %s\n", modenum, modestr);

	vga_setmode(modenum);

	modeinfo = vga_getmodeinfo(modenum);

	visinfo=VIS_PRIVATE(vis);
	
	visinfo->islinear=0;
	visinfo->ismodex=0;
	visinfo->isbanked=0;

	err=0;
	if (((modeinfo->flags & CAPABLE_LINEAR)
		&& (err=vga_setlinearaddressing()) >= tm->virt.x*tm->virt.y)) {
		visinfo->islinear=1;
  	} else {
		if ((modeinfo->flags & IS_MODEX)) {
			visinfo->ismodex=1;
		} else {
			if (modeinfo->linewidth*tm->virt.y <= 1<<16)
				visinfo->islinear=1;
			else {
				visinfo->isbanked=1;
			}
		}
	}
	
	if (visinfo->islinear) {
		VIS_FB_LINEAR(vis)=vga_getgraphmem();
		VIS_FB_LINEAR_SIZE(vis) = modeinfo->linewidth * tm->virt.y;
	}

/* virt.x != visible.x should be possible with this,
   but currently it doesn't work. */
#if 0
	vga_setlogicalwidth((tm->virt.x*depth)/8);
#endif

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);
		
	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-SVGAlib: Can't load the \"generic-stubs\" "
				"library\n");
	}

	if (!visinfo->ismodex) {
		char sugname[80];
		char arg[80];
		*arg = '\0';
		if (visinfo->islinear) {
			sprintf(sugname,"generic-linear-%d",depth);
		} else if (visinfo->isbanked) {
			sprintf(sugname,"helper-vgagl");
			sprintf(arg,"sVgALIb");
		}
		err=(_ggiOpenDL(vis,sugname,arg)==NULL);
		if (err) {
			fprintf(stderr,"display-SVGAlib: Can't find an appropriate "
					"library for %s\n",
					sugname);
			return err;
		} else {
			DPRINT("Success in loading %s\n",sugname);
		}
	}

	/* Override the color operations for X if we are 8 bit palette. */
	if (SVGA_PRIV(vis)->colormap) {
		free(SVGA_PRIV(vis)->colormap);SVGA_PRIV(vis)->colormap=NULL;
	}

	if (VIS_MODE(vis)->graphtype==GT_8BIT) {

		int x;
		SVGA_PRIV(vis)->nocols=256;
		SVGA_PRIV(vis)->colormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<SVGA_PRIV(vis)->nocols;x++)
			{
				SVGA_PRIV(vis)->colormap[x].r=((x&0xe0)>>5)*0x2492;
				SVGA_PRIV(vis)->colormap[x].g=((x&0x1c)>>2)*0x2492;
				SVGA_PRIV(vis)->colormap[x].b=((x&0x03)   )*0x5555;
			}
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->unmappixel=GGIunmappixel;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
		/* This is just to trigger the initial setup. */
		GGIsetpalvec(vis,0,256,SVGA_PRIV(vis)->colormap);
	}

#if defined(__linux__) && defined (WANT_DANGEROUS_SPLITLINE)
	{
#define	MISCr	0x03CC
#define	MISCw	0x03C2
#define	CRTIm	0x03B4
#define CRTIc	0x03D4
#define CRTDm	0x03B5
#define	CRTDc	0x03D5
#define	MISC_COLOR_IO	0x01
		int tmp = inb(MISCr);
		outb(tmp, MISCw);
		if (tmp & MISC_COLOR_IO) { 
			visinfo->CRTI = CRTIc;
			visinfo->CRTD = CRTDc; 
		} else {
			visinfo->CRTI = CRTIm;
			visinfo->CRTD = CRTDm;
		}
	}
	vis->op2d->setsplitline=GGIsetsplitline;
#endif
	/* Doesn't work correct */
	/*vis->op2d->setorigin=GGIsetorigin;*/
	
	vis->op2d->waitraypos=GGIwaitraypos;
	if (visinfo->ismodex) {
		vis->op2d->drawpixel=GGIdrawpixel;
		vis->op2d->putpixel=GGIputpixel;
		vis->op2d->getpixel=GGIgetpixel;
		vis->op2d->_drawpixel=_GGIdrawpixel;
		vis->op2d->drawhline=GGIdrawhline;
		vis->op2d->_drawhline=_GGIdrawhline;
		vis->op2d->drawvline=GGIdrawvline;
		vis->op2d->_drawvline=_GGIdrawvline;
		vis->op2d->drawbox=GGIdrawbox;
		vis->op2d->puthline=GGIputhline;
		vis->op2d->gethline=GGIgethline;
		vis->op2d->putbox=GGIputbox;
	}
	return EOK;
}


static int _GGIcheckonebpp(ggi_visual *vis,ggi_mode *tm, svga_ggimode *gm)
{
	int i=0, bestw=0, besth=0;
	int wantw=tm->visible.x, wanth=tm->visible.y;
	for (i=0; gm[i].bpp != 0; i++) {
		if (gm[i].gt != tm->graphtype)
			continue;
		if ((gm[i].w == wantw || wantw == GGI_AUTO) &&
		    (gm[i].h == wanth || wanth == GGI_AUTO)) {
			tm->visible.x=gm[i].w;
			tm->visible.y=gm[i].h;
			return 0;
		}
		if (((gm[i].w >= bestw &&
		      bestw <= wantw) || wantw == GGI_AUTO) &&
		    ((gm[i].h >= besth &&
		      besth <= wanth) || wanth == GGI_AUTO)) {
			bestw = gm[i].w;
			besth = gm[i].h;
		}
	}
	if (bestw == 0)
		return 1;
	tm->visible.x=bestw;
	tm->visible.y=besth;
	return -1;
}

static int _GGIgethighbpp(ggi_visual *vis,ggi_mode *tm, svga_ggimode *gm)
{
	int i=0, bestw=0, besth=0, bestbpp=0;
	int wantw=tm->visible.x, wanth=tm->visible.y,
		wantbpp=ggiGT2BPP(tm->graphtype);
	for (i=0; gm[i].bpp != 0; i++) {
		if (((gm[i].w >= bestw &&
		      bestw <= wantw) || wantw == GGI_AUTO) &&
		    ((gm[i].h >= besth &&
		      besth <= wanth) || wanth == GGI_AUTO) &&
		     (gm[i].bpp >= bestbpp && bestbpp <= wantbpp)) {
			bestw = gm[i].w;
			besth = gm[i].h;
			bestbpp = gm[i].bpp;
			continue;
		}
	}
	tm->visible.x=bestw;
	tm->visible.y=besth;
	tm->graphtype=ggiBPP2GT(bestbpp);
	
	return 0;

}

static int _GGIcheckautobpp(ggi_visual *vis,ggi_mode *tm, svga_ggimode *gm)
{
	int i=0, ret=0, bestw=0, besth=0, bestbpp=0, gotmode=0;
	int wantw=tm->visible.x, wanth=tm->visible.y;
	for (i=0; gm[i].bpp != 0; i++) {
		if ((gm[i].w == wantw || wantw == GGI_AUTO) &&
		    (gm[i].h == wanth || wanth == GGI_AUTO)) {
			if (gm[i].bpp >= bestbpp) {
				bestw = gm[i].w;
				besth = gm[i].h;
				bestbpp = gm[i].bpp;
				gotmode=1;
			}
		}
		if (gotmode)
			continue;
		if (((gm[i].w >= bestw &&
		     bestw <= wantw) || wantw == GGI_AUTO) &&
		    ((gm[i].h >= besth &&
		      besth <= wanth) || wanth == GGI_AUTO) &&
		    gm[i].bpp >= bestbpp) {
			bestw = gm[i].w;
			besth = gm[i].h;
			bestbpp = gm[i].bpp;
		}
	}
	if (gotmode) {
		ret = 0;
	} else
		ret = -1;
	tm->visible.x=bestw;
	tm->visible.y=besth;
	tm->graphtype=ggiBPP2GT(bestbpp);
	
	return ret;
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	int ret, err = 0;

	if (vis==NULL && tm==NULL)
		return -1;
	
	if (tm->visible.x == GGI_AUTO &&
	    tm->virt.x != GGI_AUTO)
		tm->visible.x = tm->virt.x;
	if (tm->visible.y == GGI_AUTO &&
	    tm->virt.y != GGI_AUTO)
		tm->visible.y = tm->virt.y;
	
	if (tm->graphtype == GGI_AUTO) {
		err=_GGIcheckautobpp(vis, tm, SVGA_PRIV(vis)->availmodes);
	} else if ((ret =
		    _GGIcheckonebpp(vis, tm, SVGA_PRIV(vis)->availmodes))
		   != 0) {
		err = -1;
		if (ret == 1)
			_GGIgethighbpp(vis, tm, SVGA_PRIV(vis)->availmodes);
	}
	

	if(tm->virt.x==GGI_AUTO) tm->virt.x = tm->visible.x;
	if(tm->virt.y==GGI_AUTO) tm->virt.y = tm->visible.y;

	/* SVGAlib doesn't seem to support virtual dimensions
	   Force them to be the same as visible size */
	if(tm->virt.x != tm->visible.x) {
		tm->virt.x = tm->visible.x;
		err = -1;
	}
	if (tm->virt.y != tm->visible.y) {
		tm->virt.y = tm->visible.y;
		err = -1;
	}
 
	if(	(tm->dpp.x!=1 && tm->dpp.x!=GGI_AUTO) ||
		(tm->dpp.y!=1 && tm->dpp.y!=GGI_AUTO))
		{ err = -1; }
	tm->dpp.x = tm->dpp.y = 1;

	return err;
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;

	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));
	return 0;
}
