/* 
   Graphics library for GGI. Mode management for XF86.

   Copyright (C) 1997 Steve Cheng <elmert@ipoline.com>
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1997 Andreas Beck - becka@uni-duesseldorf.de
   
   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include "internal.h"
#include <ggi/ggi_commands.h>
#include "Xvisual.h"

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	ggi_mode *mode=VIS_MODE(vis);

	if ( x<0 || x> mode->virt.x-mode->visible.x)
		return -EINVAL;

	XF86DGASetViewPort(xhook->display, xhook->screen, x, y);

	return EOK;
}

/*
** Finds the appropriate XF86VidMode modeline for x/y.
*/
static int _GGIfindmode(ggi_visual *vis, int visible_x, int visible_y)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	int i;

	/* Find a suitable XF86VidMode and return it */
	for(i = 0; i<xhook->num_modes; i++)
		if(	visible_x == xhook->modes[i]->hdisplay &&
			visible_y == xhook->modes[i]->vdisplay)
			{ return i; }

	/* mode not found */
	return -1;
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	ggi_suggest sug;

	char sugs[1][80];

	switch(VIS_MODE(vis)->graphtype)
	{
	  case GT_8BIT : err=8;break;
	  case GT_15BIT: err=15;break;
	  case GT_16BIT: err=16;break;
	  case GT_24BIT: err=24;break;
	  case GT_32BIT: err=32;break;
	  default: return -EINVAL;
	}

	sprintf(sugs[0],"generic-linear-%d",err);

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-XF86DGA: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	sug.id=0;
	do {
		/* FIXME ! */
		strcpy(sug.name,sugs[sug.id]);
		strcpy(sug.args,"");

		err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
		if (err) {
			fprintf(stderr,"display-XF86DGA: Can't find an appropriate "
				       "library for %s (%s)\n",
					sug.name,sug.args);
			return err;
		} else {
			DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
		}
	} while (sug.id!=0);

	vis->op2d->setorigin=GGIsetorigin;

        /* Override the color operations for X if we are 8 bit palette. */
        if (XLIB_PRIV(vis)->colormap) {
                free(XLIB_PRIV(vis)->colormap);XLIB_PRIV(vis)->colormap=NULL;
        }

        if (VIS_MODE(vis)->graphtype==GT_8BIT) {

                int x;
                XLIB_PRIV(vis)->nocols=256;
                XLIB_PRIV(vis)->colormap=malloc(256*sizeof(ggi_color));
                for(x=0;x<XLIB_PRIV(vis)->nocols;x++)
                {
                        XLIB_PRIV(vis)->colormap[x].r=((x&0xe0)>>5)*0x2492;
                        XLIB_PRIV(vis)->colormap[x].g=((x&0x1c)>>2)*0x2492;
                        XLIB_PRIV(vis)->colormap[x].b=((x&0x03)   )*0x5555;
                }
                vis->opcolor->mapcolor=GGImapcolor;
                vis->opcolor->setpalvec=GGIsetpalvec;
                vis->opcolor->getpalvec=GGIgetpalvec;
                /* This is just to trigger the initial setup. */
                GGIsetpalvec(vis,0,256,XLIB_PRIV(vis)->colormap);
        }
        /* color */
        vis->opcolor->unmappixel=GGIunmappixel;

	return EOK;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct Xhooks *xhook;

	void *fb;
	int stride, bank_size, mem_size, flags;

	/* First, check if the mode can be set */
	if(GGIcheckmode(vis, tm))
		return -1;

	xhook=VIS_PRIVATE(vis);

	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
		XF86DGADirectVideo(xhook->display, xhook->screen, 0);
	}

	/* Set XF86VidMode */
	XF86VidModeSwitchToMode(xhook->display, xhook->screen,
		xhook->modes[
			_GGIfindmode(vis, tm->visible.x, tm->visible.y)
		]);

	GET_ROOT;
	XF86DGAGetVideo(xhook->display, xhook->screen, (char**) &fb,
		&stride, &bank_size, &mem_size);
	GIVE_UP_ROOT;

	DPRINT("fb: %p, stride: %d, bank_size: %d, mem_size: %d\n",
		fb, stride, bank_size, mem_size);

	if (bank_size!=mem_size*1024) {
		fprintf(stderr, "Sorry, banked framebuffer layout not supported.\n");
		return GGI_DL_ERROR;
	}
	/* FIXME: Stride is ignored. Need DirectBuffer... */

	VIS_FB_LINEAR(vis) = fb;
	VIS_FB_LINEAR_SIZE(vis) = bank_size;

/* 
 * Not needed. We already have the information.
 *
 * However, for the paranoid, the fields in xhook can be checked 
 * against the information here, and panic if they don't match.
 */
#if 0
	XF86DGAGetViewPortSize(xhook->display, xhook->screen, 
		(int*) &VIS_MODE(vis)->visible.x, (int*) &VIS_MODE(vis)->visible.y);
	DPRINT("viewport: %dx%d\n", VIS_MODE(vis)->visible.x, VIS_MODE(vis)->visible.y);
#endif	
	
	/* Get all the events */
	XGrabKeyboard(xhook->display, DefaultRootWindow(xhook->display), 
		True, GrabModeAsync, GrabModeAsync, CurrentTime);

	XGrabPointer(xhook->display, DefaultRootWindow(xhook->display),
		True, PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
		GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

	if (xhook->cmap)
		XFreeColormap(xhook->display,xhook->cmap);

	if (tm->graphtype==GT_8BIT)
	{
		XVisualInfo vinfo;
		XMatchVisualInfo (xhook->display, xhook->screen, 8, PseudoColor, &vinfo);

		xhook->cmap = XCreateColormap(xhook->display,
				DefaultRootWindow(xhook->display),
				vinfo.visual, AllocAll);

		DPRINT("8-bit visual: X-lib colormap allocated %x.\n",xhook->cmap);
	} else {
		xhook->cmap=0;
	}

	DPRINT("Ready, now taking crash course!\n"); 

	flags = XF86DGADirectGraphics | XF86DGADirectMouse | XF86DGADirectKeyb;
	XF86DGADirectVideo(xhook->display, xhook->screen, flags);

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return _GGIdomode(vis);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	struct Xhooks *xhook;
	int depth;
	int err = 0;	/* Flag: if the mode being checked is not found */

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}

	xhook=VIS_PRIVATE(vis);

	if(tm->visible.x==GGI_AUTO) tm->visible.x = xhook->modes[0]->hdisplay;
	if(tm->visible.y==GGI_AUTO) tm->visible.y = xhook->modes[0]->vdisplay;
	if(tm->virt.x==GGI_AUTO) tm->virt.x = xhook->width;
	if(tm->virt.y==GGI_AUTO) tm->virt.y = xhook->height;

/*
	if(	(tm->dpp.x!=1 && tm->dpp.x!=GGI_AUTO) ||
		(tm->dpp.y!=1 && tm->dpp.y!=GGI_AUTO))
		{ err = -1; }
	
	dpp.x = 1;
	dpp.y = 1;
*/

	switch(tm->graphtype)
	{
	  case GT_8BIT : depth=8;break;
	  case GT_15BIT: depth=15;break;
	  case GT_16BIT: depth=16;break;
	  case GT_24BIT: depth=24;break;
	  case GT_32BIT: depth=32;break;
	  default: depth=0;
	}
	if(depth != xhook->depth) {
		if(tm->graphtype != GGI_AUTO)
			err = -1;
		switch(xhook->depth)
		{
		  case 8 : tm->graphtype=GT_8BIT; break;
		  case 15: tm->graphtype=GT_15BIT; break;
		  case 16: tm->graphtype=GT_16BIT; break;
		  case 24: tm->graphtype=GT_24BIT; break;
		  case 32: tm->graphtype=GT_32BIT; break;
		  ggiPanic("Current bitdepth is unsupported by LibGGI!\n");
		}
	}

	/* Virtual resolution is unchangeable */
	if(tm->virt.x != xhook->width || tm->virt.y != xhook->height) {
		tm->virt.x = xhook->width;
		tm->virt.y = xhook->height;
		err = -1;
	}
	
	/* Try to find a suitable XF86 mode. */
	if(_GGIfindmode(vis, tm->visible.x, tm->visible.y)==-1) {
		int sug_x, sug_y, new_x, new_y, i;

		/* If that cannot be done, flag it as an error */
		err = -1;

		/* Then try to find another mode with
		   the next higher resolution */

		sug_x = sug_y = 0x7FFFFFFF;	/* Just the maximum */

		for(i = 0; i<xhook->num_modes; i++) {
			new_x = xhook->modes[i]->hdisplay;
			new_y = xhook->modes[i]->vdisplay;
			if(	new_x >= tm->visible.x &&
				new_y >= tm->visible.y &&
				new_x <= sug_x &&
				new_y <= sug_y)
			{
				sug_x = new_x;
				sug_y = new_y;
			}
		}
		
		/* If the above failed, then use
		   the highest resolution possible */
		if(	sug_x == 0x7FFFFFFF &&
			sug_y == 0x7FFFFFFF)
		{
			sug_x = sug_y = 0;

			for(i = 0; i<xhook->num_modes; i++) {
				new_x = xhook->modes[i]->hdisplay;
				new_y = xhook->modes[i]->vdisplay;
				if(new_x >= sug_x && new_y >= sug_y) {
					sug_x = new_x;
					sug_y = new_y;
				}
			}
		}
		
		tm->visible.x = sug_x;
		tm->visible.y = sug_y;
	}			

	return err;	
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
	if (VIS_MODE(vis)==NULL)
		return -1;

	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));

	return EOK;
}
