/*
  !!! WARNING: This file is for libggi 1.4 compatibility
*/

/* 
 *  Library initialization
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "internal.h"

/*
 * Initalize the strutures for the library
 */
static int 	_ggi2dLibIsUp=0;

int ggi2dDebugState=0;
int ggi2d_offset;

/* Routine for loading the extension libraries
 */
static int ggi2dLoad(ggi_visual_t vis,const char *driver,const char *args)
{
	char str[1024];
        
	strcpy(str,driver);
	strcat(str,"-2d");
                        
	if (ggiLoadLibrary(vis,str,args)) {
		DPRINT("vis %p:  %s loaded!\n",vis,str);
	}
	else
	{
		if (strcmp(driver,"generic-stubs")==0)
		{	/* If this ain't there, there is something really wrong ... */
			DPRINT("%s not found !\n",str);
			DPRINT("Please check your config files!\n");
		}
	}
	return 0;
}

static int ggi2dRelease(ggi_visual_t vis,const char *driver)
{
        /* ggiCloseLibrary()  FIXME ! */
	return 0;
}

int ggi2dInit(void)
{
	const char *str;

	_ggi2dLibIsUp++;
	if (_ggi2dLibIsUp!=1) return EOK;	/* Initialize only at first call. */

	str=getenv("LIBGGI2D_DEBUG");
	if (str!=NULL) {
		ggi2dDebugState=atoi(str);
		DPRINT("Debugging=%d\n",ggi2dDebugState);
	}

	ggiInit();
	DPRINT("LibGGI2D initialized.\n");

        ggi2d_offset=ggiRegisterExtension(sizeof(ggi2d_data),
                                        ggi2dLoad,ggi2dRelease);

	DPRINT("Registered Extension at %d.\n",ggi2d_offset);
                                        
	return EOK;
}

void ggi2dExit(void)
{
	if (!_ggi2dLibIsUp)
		return;

	if (_ggi2dLibIsUp!=1) return;	/* Exit only at last call */

	DPRINT("ggi2dExit: really destroying.\n");
	
	ggiExit();
	_ggi2dLibIsUp--;
	return;
}

/*----------------------------------------------------------------------*/

int	ggi2dOpen(ggi_visual_t vis)
{
	DPRINT("ggi2dOpen: opening visual.\n");
	GGI2DinitGC(vis);    
	GGI2Dupdate(vis, GGI2D_UPD_ALL);    
	return EOK;
}

void	ggi2dClose(ggi_visual_t vis)
{
	DPRINT("ggi2dClose: closing visual.\n");
}

