/*
 *	Copyright (C) 1998 Uwe Maurer - uwe_maurer@t-online.de
 */

#include <ggi/libggu.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

ggi_color pal[256];
ggi_color pal1[256];
ggi_color pal2[256];

ggi_visual_t vis;
ggi_visual_t vismem;
ggi_visual_t image;

#define NUM	8	

int xpos,ypos,width,height;

void calc_palette(void)
{
	int i,j,o;
	ggi_color black[256];
	memset(black,0,sizeof(black));
	ggiSetPaletteVec(vis,0,256,black);

	for (i=0;i<NUM;i++) 
	{
		pal1[i]=pal2[i]=pal[i];
	}

	for (o=NUM,i=0;i<NUM;i++)
	{
		for (j=0;j<NUM;j++,o++)
		{
			pal1[o]=pal[j];
			pal2[o]=pal[i];
		}
	}
}
void calc_next(ggi_visual_t vis1,ggi_visual_t vis2)
{
	int x,y;
	ggi_pixel p,p1,p2;

	for (y=ypos;y<ypos+height;y++)
	for (x=xpos;x<xpos+width;x++)
	{
		ggiGetPixel(vis1,x,y,&p1);
		ggiGetPixel(vis2,x,y,&p2);
		p=NUM+p1+(p2*NUM);
		ggiPutPixel(vis1,x,y,p);		
	}
}

void reset(ggi_visual_t vis)
{
	int x,y;
	ggi_pixel p;

	for (y=ypos;y<ypos+height;y++)
	for (x=xpos;x<xpos+width;x++)
	{
		ggiGetPixel(vis,x,y,&p);
		p=(p-NUM)/NUM;
		ggiPutPixel(vis,x,y,p);		
	}
}

void fade(void)
{
	int y;
	char *linear=gguGetPLB(image)->read;

	if (!linear)	
	{
		ggiPanic("Can't get pixellinearbuffer !\n");
	}
	for (y=0;y<6*height;y+=height)
	{
		ggiPutBox(vismem,xpos,ypos,width,height,linear+y*width);
		calc_next(vis,vismem);
		gguPalFadeTo(vis,pal2);
		reset(vis);	
		ggiSetPaletteVec(vis,0,256,pal1);
	}
	gguPalFadeOut(vis);
	ggiSetGCForeground(vis,0);
	ggiFillscreen(vis);
}

void part2(void)
{	
	#define START	16	
	#define LEN	20
	#define COLORS  8
	#define END	(START + COLORS * LEN)
	
	ggi_color col[COLORS];
	int i;
	float x=(float)width/2,y=(float)height/2,w1,r1;
	float w2,r2;
	float x2,y2;
	int f,f2=START;

	col[0]=gguBlack;
	col[1]=gguBlue;
	col[2]=gguRed;
	col[3]=gguYellow;
	col[4]=gguColor(.8,.2,.6);
	col[5]=gguColor(.6,.2,.8);
	col[6]=gguBlack;
	col[7]=gguWhite;

	for (i=0;i<COLORS;i++)
		gguPalBlend(pal1,col[i],col[(i+1)%COLORS],START+i*LEN,LEN);

	memset(pal2,0,sizeof(pal2));
	for (i=0;i<COLORS;i++)
		gguPalBlend(pal2,col[i],gguBlack,START+LEN*i,LEN/2);

	ggiSetPaletteVec(vis,0,256,pal1);

	r1=2.0/3.0*y;
	r2=1.0/3.0*y;

	for (;r1>0;r1-=5,r2-=r1/26)
	{	
		f2+=8;
		if (f2>=END) f2-=(END-START);	
		f=f2;
		for (w1=0,w2=0;w1<2*M_PI;w1+=M_PI/512,w2+=M_PI/64)
		{
			x2=x+sin(w1)*r1+sin(w2)*r2;	
			y2=y+cos(w1)*r1+cos(w2)*r2;	
			if (f>=END) f=START;
			ggiPutPixel(vis,x2,y2,f++);
		}
	}

#if 0
	for (x=0;x<256;x++)
	for (y=0;y<25;y++)
		ggiPutPixel(vis,x,y,x);
#endif

	gguPalRange(vis,START,LEN*COLORS);
	while (!ggiKbhit(vis))
	{
		gguPalMove(vis,-1);
	} ggiGetc(vis);
	gguPalFadeTo(vis,pal2);
	while (!ggiKbhit(vis))
	{
		gguPalMove(vis,2);
	} ggiGetc(vis);
}

int main(int argc,char *argv[])
{
	gguInit();

	vis=gguCmdLineVisual(&argc,&argv,"--ggimode 320x200x8");

	if (gguBPP(GGU_GT(vis))!=8) 
	{
		printf("Only 8-bpp modes please!\n");
		exit(1);
	}

	vismem=ggiOpen("display-memory",NULL);
	image=ggiOpen("display-memory",NULL);

	if(gguLoadImage(image,"pal.gif",pal))
	{
		ggiPanic("can't find pal.gif\n");
	}

	ggiSetMode(vismem,GGU_MODE(vis));

	gguPalFadeSteps(vis,15);

	calc_palette();
	width=GGU_WIDTH(image);
	height=30;
	xpos= (GGU_WIDTH(vis)-width)/2;
	ypos= (GGU_HEIGHT(vis)-height)/2;
	fade();	
	width=GGU_WIDTH(vis);
	height=GGU_HEIGHT(vis);
	part2();
	ggiClose(vis);
	ggiClose(vismem);
	ggiClose(image);
	gguExit();
	exit(0);
}
