/* hhanemaa@cs.ruu.nl */
/* Mouse library functions */

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include "vga.h"
#include "vgamouse.h"

static void default_handler(int button, int dx, int dy, int dz, int drx,
                            int dry, int drz);
static void (*__mouse_eventhandler) (int, int, int, int, int, int, int) = default_handler;

#include "ms.inc"			/* include low-level mouse driver */

int mouse_init_return_fd(char *dev, int type, int samplerate)
{
    /* Install default mouse handler. Old coordinates are preserved. */
    __mouse_eventhandler = default_handler;

    return 0;/*EVENT_FD;*/			/* Return mouse fd. */
}

/* Old compatible mouse_init, returns 0 if succesful, -1 on error. */
int mouse_init(char *dev, int type, int samplerate)
{
   mouse_init_return_fd(NULL,0,0);
	return 0;
}

/* For now, we assume there's no console switching. */

int mouse_update()
{
    int result;
    result = get_ms_event(0);
    return result;
}

void mouse_waitforupdate()
{
    get_ms_event(1);
    return;
}

void mouse_seteventhandler(__mouse_handler handler)
{
    __mouse_eventhandler = handler;
}

void mouse_close()
{
}


/* Default event handler. */

void mouse_setdefaulteventhandler()
{
    __mouse_eventhandler = default_handler;
}

static int mouse_x = 0;
static int mouse_y = 0;
static int mouse_button = 0;
static int scale = 1;
static int minx = 0;
static int maxx = 32767;
static int miny = 0;
static int maxy = 32767;
static int wrap = 0;

void mouse_setwrap(int state)
{
    wrap = state;
}

static void default_handler(int button, int dx, int dy, int dz, int drx,
                            int dry, int drz)
{
    mouse_button = button;
    mouse_x += dx;
    mouse_y += dy;
    if (mouse_x / scale > maxx)
	if (wrap & MOUSE_WRAPX)
	    mouse_x -= (maxx - minx) * scale;
	else
	    mouse_x = maxx * scale;
    if (mouse_x / scale < minx)
	if (wrap & MOUSE_WRAPY)
	    mouse_x += (maxx - minx) * scale;
	else
	    mouse_x = minx * scale;
    if (mouse_y / scale > maxy)
	if (wrap & MOUSE_WRAPX)
	    mouse_y -= (maxy - miny) * scale;
	else
	    mouse_y = maxy * scale;
    if (mouse_y / scale < miny)
	if (wrap & MOUSE_WRAPY)
	    mouse_y += (maxy - miny) * scale;
	else
	    mouse_y = miny * scale;
}

void mouse_setposition(int x, int y)
{
    mouse_x = x;
    mouse_y = y;
}

void mouse_setxrange(int x1, int x2)
{
    minx = x1;
    maxx = x2;
}

void mouse_setyrange(int y1, int y2)
{
    miny = y1;
    maxy = y2;
}

void mouse_setscale(int s)
{
    scale = s;
}

int mouse_getx()
{
    return mouse_x / scale;
}

int mouse_gety()
{
    return mouse_y / scale;
}

int mouse_getbutton()
{
    return mouse_button;
}
