#!/bin/bash
#
#	This script takes three arguments:
#	$1	path to the linux sources
#	$2	path to the ggi sources
#	$3	patch version to use
#

PATCHDIR=`pwd`/$3

echo Applying patches for version $3

if [ -d $1/include/ggi ]; then
 	echo "KGI patch already applied.";
	exit 1;
fi

echo storing backup of $1/drivers/char in $1/drivers/char.old
cp -d -R $1/drivers/char $1/drivers/char.old

echo moving obsolete/replaced files to $1/drivers/char/obsolete
mkdir $1/drivers/char/obsolete
for i in `cat $PATCHDIR/obsolete_files`; do
	mv -f $1/drivers/char/$i $1/drivers/char/obsolete
done;

echo installing KGI files to $1/driver/char
cp -d -R $PATCHDIR/../linux/drivers/char $1/drivers/
cp -d -R $PATCHDIR/../linux/include $1/

echo applying modifications to remaining files in linux/*
cd $1
patch -p1 < $PATCHDIR/mega.diff 2> mega.diff.patch.err;
if grep fail mega.diff.patch.err; then
  echo Error: the patch failed, make sure you have an unmodified kernel!
  exit -1;
fi
if grep malformed mega.diff.patch.err; then
  echo Error: malformed patch, may be the patches are broken?
  exit -1;
fi

cd $1/include && ln -s $2/include/ggi ggi
echo $1 > $2/driver/.kernelsrc

echo ""
cd $1/drivers/char/default

defmap="kbd-american.c"
map=$KEYMAP

while [ -z "$map" ]; do
	echo "Available Keymaps: "
	echo ""
	for d in kbd-*.c; do echo "	$d"; done
	echo ""
	echo -n "Keymap to use [$defmap]:"
	read map
	if [ -z "$map" ]; then
		map=$defmap
	fi
	if [ ! -f "$map" ]; then
		map=""
	fi
done
echo "Using keyboard map: $map"

rm -f keymap.c
ln -s $map keymap.c

echo ""
echo "Patches completed"

